# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/10 20:36

"""

from Bio.Seq import Seq
from Bio import SeqIO
from public_function import make_dir


class GetSeqForHmmer(object):
    def __init__(self, blast_flank_seq, simplify_gff3,
                 duplicate_removal_gff3,
                 seq_with_ltrs_nt, seq_with_ltrs_aa,
                 flank_without_ltrs_nt, flank_without_ltrs_aa):
        """
        Make sequence for HMMER analysis
        :param blast_flank_seq: flank seq from balst
        :param simplify_gff3: simplified gff3 file
        :param duplicate_removal_gff3: simplified gff3 with duplicate removal
        """
        super(GetSeqForHmmer, self).__init__()

        self.blast_flank_seq = blast_flank_seq
        self.simplify_gff3_path = simplify_gff3
        self.duplicate_removal_path = duplicate_removal_gff3

        self.seq_with_ltrs_nt = seq_with_ltrs_nt
        self.seq_with_ltrs_aa = seq_with_ltrs_aa

        self.flank_without_ltrs_nt = flank_without_ltrs_nt
        self.flank_without_ltrs_aa = flank_without_ltrs_aa

    def run(self):


        flank_seq_dict = SeqIO.to_dict(SeqIO.parse(self.blast_flank_seq, "fasta"))

        duplicate_removal_file = open(self.duplicate_removal_path,
                                      "r",encoding="utf-8")
        

        # 1. Extracted ERVs with paired LTR
        pairLTR_eves_file = open(self.seq_with_ltrs_nt, "w",
                                 encoding="utf-8")
        pairLTR_evesAA_file = open(self.seq_with_ltrs_aa, "w",
                                   encoding="utf-8")

        for line in duplicate_removal_file:
            if not line.startswith("#"):
                ltr_record_list = line.strip().split("\t")
                ltr1_start = int(ltr_record_list[3].split("..")[0])
                ltr2_end = int(ltr_record_list[4].split("..")[1])

                qury_erv_start = ltr1_start
                qury_erv_end = ltr2_end
        
                erv_seq_contain = Seq(
                    str(flank_seq_dict[ltr_record_list[0]][
                        qury_erv_start - 1: qury_erv_end].seq))

                pairLTR_eves_file.write(">" + ltr_record_list[-3] + "\n"
                                        + str(erv_seq_contain) + "\n")
        
        
                pairLTR_evesAA_file.write(">" + ltr_record_list[-3]
                                          + "#ORF1" + "\n"
                                          + str(
                    erv_seq_contain.translate(table=1)) + "\n")
        
                pairLTR_evesAA_file.write(">" + ltr_record_list[-3]
                                          + "#ORF2" + "\n"
                                          + str(
                    erv_seq_contain[1:].translate(table=1))
                                          + "\n")
        
                pairLTR_evesAA_file.write(">" + ltr_record_list[-3]
                                          + "#ORF3" + "\n"
                                          + str(
                    erv_seq_contain[2:].translate(table=1))
                                          + "\n")
        

        
        duplicate_removal_file.close()
        pairLTR_eves_file.close()
        pairLTR_evesAA_file.close()


        # 2. Extract flank sequences without paired LTR
        # using the harvest_gff3_simplify.txt
        gff3_file = open(self.simplify_gff3_path, "r",encoding="utf-8")
        
        no_pairLTR_seq = open(self.flank_without_ltrs_nt, "w",encoding="utf-8")
        no_pairLTR_seqAA = open(self.flank_without_ltrs_aa, "w",encoding="utf-8")
        
        ltr_envs_record = set()
        
        for line in gff3_file:
            ltr_envs_record.add(line.split("\t")[0].strip())
        
        for key in flank_seq_dict:
            if key.strip() not in ltr_envs_record:
                seq_contain = Seq(str(flank_seq_dict[key].seq))
        
                no_pairLTR_seq.write(">" + key + "\n" + str(seq_contain) + "\n")
        

                no_pairLTR_seqAA.write(">" + key + "#ORF1" + "\n"
                                       + str(seq_contain.translate(table=1))
                                       + "\n")
        
                no_pairLTR_seqAA.write(">" + key + "#ORF2" + "\n"
                                       + str(seq_contain[1:].translate(table=1))
                                       + "\n")
        
                no_pairLTR_seqAA.write(">" + key + "#ORF3" + "\n"
                                       + str(seq_contain[2:].translate(table=1))
                                       + "\n")
        


        
        no_pairLTR_seq.close()
        no_pairLTR_seqAA.close()



def GetSeqForHmmer2(input_file, outnt, outAA):
    flank_seq_dict = SeqIO.to_dict(SeqIO.parse(input_file, "fasta"))
    no_pairLTR_seq = open(outnt, "w", encoding="utf-8")
    no_pairLTR_seqAA = open(outAA, "w", encoding="utf-8")

    for key in flank_seq_dict:
        seq_contain = Seq(str(flank_seq_dict[key].seq))

        no_pairLTR_seq.write(">" + key + "\n" + str(seq_contain) + "\n")

        no_pairLTR_seqAA.write(">" + key + "#ORF1" + "\n"
                               + str(seq_contain.translate(table=1))
                               + "\n")

        no_pairLTR_seqAA.write(">" + key + "#ORF2" + "\n"
                               + str(seq_contain[1:].translate(table=1))
                               + "\n")

        no_pairLTR_seqAA.write(">" + key + "#ORF3" + "\n"
                               + str(seq_contain[2:].translate(table=1))
                               + "\n")


    no_pairLTR_seq.close()
    no_pairLTR_seqAA.close()


# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/10 14:25

"""
from Bio import SeqIO
from public_function import reversecomp_seq,get_all_path

class GetFlanks(object):

    def __init__(self, host_seq, blast_out_dir,
                 flanks_lenth, out_path):
        """
        Get flanks sequence near the blast hit sites
        :param host_seq: path of input genome or sequence
        :param blast_out_dir: path of blast_result
        :param flanks_lenth: the lenth of flanks_lenth
        :param out_path: path flanks sequence
        """
        super(GetFlanks, self).__init__()

        self.host_path = host_seq
        self.blast_out_dir = blast_out_dir # Support for multiple probes
        self.out_path = out_path
        self.flank_length= flanks_lenth


    def run(self):
        blast_file_list = get_all_path(self.blast_out_dir)

        EVEs = {}  # to combine the blast results of protein probes

        for each_blast_path in blast_file_list:
            matchlenth_threshold = 0
            each_blast_path = each_blast_path.strip()

            if each_blast_path.upper().find("RT") != -1:
                matchlenth_threshold = 50   # alignment length of aa

            elif each_blast_path.upper().find("GAG") != -1:
                matchlenth_threshold = 150  # alignment length of aa

            elif each_blast_path.upper().find("ENV") != -1:
                matchlenth_threshold = 200  # alignment length of aa


            blast_file = open(each_blast_path,"r",encoding="utf-8")

            for line in blast_file:

                blast_line_list = line.strip().split("\t")

                if int(blast_line_list[3]) > matchlenth_threshold:
                    if int(blast_line_list[-4]) > int(blast_line_list[-3]):

                        # On the complementary strand of the input genome sequence
                        strand = "complement"

                        """
                        the starting position is relabeled according to the 
                        numerical value from small to large
                        """
                        start = int(blast_line_list[-3])
                        end = int(blast_line_list[-4])
                    else:
                        strand = "same"  # On the same chain of the input sequence
                        start = int(blast_line_list[-4])
                        end = int(blast_line_list[-3])

                    new_line_list = [start, end, strand,
                                     float(blast_line_list[2]),
                                     int(blast_line_list[3]),
                                     float(blast_line_list[-2])]

                    # start, end, strand, pident, lenth, e-value

                    if not EVEs.__contains__(blast_line_list[1]):
                        EVEs[blast_line_list[1]] = []

                    # Add RT protein comment
                    EVEs[blast_line_list[1]].append(new_line_list)

            blast_file.close()

        ref_EVEs = {}    # processing hit event of continuous blast sites

        for subject_name in EVEs.keys():

            hit_list = EVEs[subject_name]

            # Sort the nested list according to the starting position
            hit_list_sort = sorted(hit_list,
                                   key=lambda x: x[0])

            # Sort the nested list according to the direction of the chain
            hit_list_sort = sorted(hit_list_sort,
                                   key=lambda x: x[2])
            # print(subject_name,hit_list_sort)

            # the first hit in the list as the starting value
            last_start = hit_list_sort[0][0]
            last_end = hit_list_sort[0][1]
            last_strand = hit_list_sort[0][2]

            fragment_distance = 10000  # distance threshold for blast consecutive hit events

            if not ref_EVEs.__contains__(subject_name):
                ref_EVEs[subject_name] = []

            """
            A more ingenious method is used here. After exiting the continuous 
            hit area, the range of the last continuous hit area is recorded. 
            """
            for each_hit in hit_list_sort:

                # If two blast hits on the same chain
                if each_hit[2] == last_strand:

                    # Judging the starting point distance
                    if each_hit[0] <= last_start + fragment_distance:

                        # Take the smaller value as the starting point
                        last_start = last_start

                        # Take the larger value as the end point
                        last_end = max(last_end, each_hit[1])

                    # if out of range
                    else:
                        # Record the result of the previous
                        ref_EVEs[subject_name].append(
                            [last_start, last_end, last_strand])

                        last_start = each_hit[0]
                        last_end = each_hit[1]
                        last_strand = each_hit[2]

                # If two blast hits on different chains
                else:
                    # Record the last result
                    ref_EVEs[subject_name].append(
                        [last_start, last_end, last_strand])

                    last_start = each_hit[0]
                    last_end = each_hit[1]
                    last_strand = each_hit[2]

            # Supplement the last result
            ref_EVEs[subject_name].append([last_start, last_end, last_strand])

        host_seq_dict = SeqIO.to_dict(SeqIO.parse(self.host_path, "fasta"))
        """
        Note: the key value in host_seq_dict only contains the part before
        the space in the sequence name 
        """

        # self.flank_length = 15000
        seqs = {}
        sum_lsit = []

        for subject_name in ref_EVEs.keys():
            for a in ref_EVEs[subject_name]:
                sub_id = subject_name
                s_start = a[0]
                s_end = a[1]
                strand = a[2]

                start = int(s_start) - int(self.flank_length)
                end = int(s_end) + int(self.flank_length)

                new_seq = ""

                if start <= 1 and end <= len(host_seq_dict[sub_id].seq[:]):
                    start = 1
                    end = end
                    new_seq = host_seq_dict[sub_id].seq[start - 1:end]

                elif start >= 1 and end <= len(host_seq_dict[sub_id].seq[:]):
                    start = start
                    end = end
                    new_seq = host_seq_dict[sub_id].seq[start - 1: end]

                elif start >= 1 and end >= len(host_seq_dict[sub_id].seq[:]):
                    start = start
                    end = len(host_seq_dict[sub_id].seq[:])
                    new_seq = host_seq_dict[sub_id].seq[start - 1: end]

                elif start <= 1 and end >= len(host_seq_dict[sub_id].seq[:]):
                    start = 1
                    end = len(host_seq_dict[sub_id].seq[:])
                    new_seq = host_seq_dict[sub_id].seq[start - 1:end]

                else:
                    print("error")

                name = (">" + sub_id + "|" + str(start) + "|" + str(end) + "|"
                        + strand)
                seq = (str(new_seq))

                seqs[name] = seq

                sum_lsit.append(name)

        #  Remove areas with full containment relations (leave wide areas) in same chain
        use_list = []

        for line1 in sum_lsit:
            qury_name, qury_start, qury_end, qury_direction = line1.split("|",
                                                                          3)
            for line2 in sum_lsit:

                if line1 != line2:

                    refer_name, refer_start, refer_end, refer_direction = line2.split(
                        "|", 3)
                    if (qury_name == refer_name and
                            qury_direction == refer_direction):
                        if qury_start >= refer_start and qury_end <= refer_end:
                            break

            else:
                use_list.append(line1)



        ganjing = set()

        finnaly_use = []

        for each_seqname in use_list:
            each_seqname = each_seqname.strip()

            if not each_seqname in ganjing:
                finnaly_use.append(each_seqname)
                ganjing.add(each_seqname)

        # get flanking sequence
        with open(self.out_path, "w",encoding="utf-8") as erv_flanking:
            for line in finnaly_use:

                if line.split("|")[-1] == "complement":
                    erv_flanking.write(line + "\n"
                                        + reversecomp_seq(
                        seqs[line].strip()) + "\n")   #  Reverse complement the sequence on the negative strand
                else:
                    erv_flanking.write(line + "\n" + seqs[line].strip() + "\n")





# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/10 20:46

"""

import os

from public_function import read_gff3

class SimplyGFF(object):

    def __init__(self, input_gff3,
                 simplify_gff3,
                 duplicate_removal_gff3,
                 mintsd):
        super(SimplyGFF, self).__init__()
        
        self.gff3_path = input_gff3
        self.simplify_gff3_path = simplify_gff3
        self.duplicate_removal_path = duplicate_removal_gff3
        self.mintsd = int(mintsd)

    def run(self):

        record_dict = {}

        seqs, annotations = read_gff3(self.gff3_path)


        for key in annotations.keys():
            annos = annotations[key]

            name = ""

            for line in annos:
                anno_list = line.strip().split("\t")

                if anno_list[2] == "repeat_region":
                    region = anno_list[3] + ".." + anno_list[4]
                    name = anno_list[0] + "|" + region
                    if not record_dict.__contains__(name):
                        record_dict[name] = []

                elif anno_list[2] == "LTR_retrotransposon":
                    similarity = anno_list[-1].split(";")[-2].replace("ltr_similarity=", "")
                    record_dict[name].append(similarity)

                elif anno_list[2] == "long_terminal_repeat":
                    ltr_region = anno_list[3] + ".." + anno_list[4]
                    record_dict[name].append(ltr_region)


        # print(record_dict)

        out_simplify_file = open(self.simplify_gff3_path, "w",encoding="utf-8")
        out_simplify_file.write(
            "#name of flank seq" + "\t" + "repeat_region in flank seq"
            + "\t" + "similarity of pairLTRs" + "\t" + "LTR1 in flank seq"
            + "\t" + "LTR2 in flank seq" + "\t" * 2
            + "seq_name_reset(map to the genome)" + "\t"
            + "LTR1 in genome" + "\t" + "LTR2 in genome" + "\n")

        for k, y in record_dict.items():

            qury_seq_name = ""

            if seqs == {}:
                LTRharvest_out_dir = os.path.dirname(self.gff3_path)

                LTRharvest_out_file = LTRharvest_out_dir + "/" + "LTRharvest_out.fas"

                with open(LTRharvest_out_file, "r", encoding='utf-8') as f:

                    for line in f:
                        if line.startswith(">"):
                            qury_seq_name = line.split("(dbseq")[0].strip().strip(">")

            else:
                qury_seq_name = seqs[k.split("|")[0]]  #  name of query sequence

            repeat_region_range = k.split("|")[1].strip()

            ltr1_range = y[1]  # range of ltr1
            ltr2_range = y[2]  # range of ltr2

            # print(y)

            """
            Recalculate the start and end positions of the endogenous 
            viral element on the original genome for renaming the sequence name 
            """

            # The start position of the flank sequence on the input genome
            start_nt = int(qury_seq_name.split("|")[1])

            # The end position of the flank sequence on the input genome
            end_nt = int(qury_seq_name.split("|")[2])
            seq_lenth = end_nt - start_nt + 1

            # start and end position in candidate pair-LTR ervs
            erv_start = int(repeat_region_range.split("..")[0])
            erv_end = int(repeat_region_range.split("..")[1])

            ltr1_start = int(repeat_region_range.split("..")[0])
            # print(ltr1_start)

            ltr1_end = int(ltr1_range.split("..")[1])

            ltr2_start = int(ltr2_range.split("..")[0])
            ltr2_end = int(repeat_region_range.split("..")[1])

            # whether it is on the complementary chain
            if qury_seq_name.split("|")[-1] == "complement":

                erv_start = seq_lenth - int(repeat_region_range.split("..")[1]) + 1
                erv_end = seq_lenth - int(repeat_region_range.split("..")[0]) + 1


                ltr1_start = seq_lenth - int(ltr1_range.split("..")[1]) + 1

                ltr1_end = seq_lenth - int(repeat_region_range.split("..")[0]) + 1

                ltr2_start = seq_lenth - int(repeat_region_range.split("..")[1]) + 1


                ltr2_end = seq_lenth - int(ltr2_range.split("..")[0]) + 1


            erv_start_site = start_nt + erv_start - 1
            erv_end_site = start_nt + erv_end - 1

            ltr1_start_site = start_nt + ltr1_start - 1
            ltr1_end_site = start_nt + ltr1_end - 1

            ltr2_start_site = start_nt + ltr2_start - 1
            ltr2_end_site = start_nt + ltr2_end - 1

            seq_name_reset = (
                        qury_seq_name.split("|")[0] + "|"
                        + str(erv_start_site) + "|"
                        + str(erv_end_site) + "|"
                        + qury_seq_name.split("|")[-1])


            out_simplify_file.write(
                qury_seq_name + "\t" + k.split("|")[1] + "\t" + y[0] + "\t"
                + str(int(repeat_region_range.split("..")[0])) + ".."
                + str(int(ltr1_range.split("..")[1])) + "\t"
                + str(int(ltr2_range.split("..")[0])) + ".."
                + str(int(repeat_region_range.split("..")[1]))

                + "\t" * 2 + seq_name_reset + "\t"
                + str(ltr1_start_site) + ".." + str(ltr1_end_site) + "\t"
                + str(ltr2_start_site) + ".." + str(ltr2_end_site) + "\n")


        out_simplify_file.close()

        out_duplicate_removal = open(self.duplicate_removal_path, "w",encoding="utf-8")

        simplify_gff3 = open(self.simplify_gff3_path, "r",encoding="utf-8")

        lines_seen = set()
        for line in simplify_gff3:
            if line.startswith("#"):
                out_duplicate_removal.write(line)
            else:
                seq_name = line.split("\t")[-3]

                if seq_name not in lines_seen:
                    out_duplicate_removal.write(line)

                    lines_seen.add(seq_name)

        simplify_gff3.close()
        out_duplicate_removal.close()






# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/10 23:03

"""
retroviridae_list = ["alpharetrovirus-related",
                     "betaretrovirus-related",
                     "gammaretrovirus-related",
                     "deltaretrovirus-related",
                     "epsilonretrovirus-related",
                     "lentivirus-related",
                     "spumaretrovirinae-related",
                     "retroviridae-related"]


from public_function import best_domain

class SimplyHMMER(object):
    def __init__(self, hmmer_file, out_simplify, header_str,para_dic):
        super(SimplyHMMER,self).__init__()

        self.hmmer_result_path = hmmer_file
        self.hmmer_simplify_path = out_simplify
        self.header_str = header_str

        self.para_dic = para_dic

        # print(self.para_dic["GAG_length"])
        #
        # print(self.para_dic["PRO_length"])
        #
        # print(self.para_dic["RT_length"])
        #
        # print(self.para_dic["RNaseH_length"])
        #
        # print(self.para_dic["INT_length"])
        #
        # print(self.para_dic["ENV_length"])


    def run(self):

        hmmer_result_file = open(self.hmmer_result_path, "r",encoding="utf-8")

        HMMER_simplify_file = open(self.hmmer_simplify_path, "w",
                                   encoding="utf-8")

        HMMER_simplify_file.write(self.header_str)

        sum_dict = {}

        for line in hmmer_result_file:

            if not line.startswith("#"):
                line_list = [i for i in line.strip().split(' ') if i != ""]

                # print(line_list)

                new_line = []

                domain = line_list[0] + "-related" # record of predicted domain

                ERV_candidate_seqname = (line_list[3].split("#")[0])

                orf = line_list[3].split("#")[-1]

                c_Evalue = float(line_list[11])

                # print(c_Evalue)

                start_aa = (int(line_list[17]))

                end_aa = (int(line_list[18]))

                domain_aa_lenth = end_aa - start_aa + 1 

                new_line.append(domain)
                new_line.append(c_Evalue)
                new_line.append(orf)
                new_line.append(
                    "AA(" + str(start_aa) + ".." + str(end_aa) + ")")
                new_line.append(domain_aa_lenth)

                if not sum_dict.__contains__(ERV_candidate_seqname):
                    sum_dict[ERV_candidate_seqname] = []
                sum_dict[ERV_candidate_seqname].append(new_line)

        for key in sum_dict: 

            line_flage = "False"

            retroviridae_flage = True

            # print(key, sum_dict[key])
            hit_list = sum_dict[key]
            GAG_list = []
            DUT_list = []
            AP_list = []
            RT_lsit = []
            RNaseH_list = []
            INT_list = []
            ENV_list = []

            for each_hit in hit_list:

                # print(each_hit)
                match_domain = each_hit[0]

                if match_domain.split("_")[0] == "GAG":
                    GAG_list.append(each_hit)

                elif match_domain.split("_")[0] == "DUT":
                    DUT_list.append(each_hit)

                elif match_domain.split("_")[0] == "AP":
                    AP_list.append(each_hit)

                elif match_domain.split("_")[0] == "RT":
                    RT_lsit.append(each_hit)

                elif match_domain.split("_")[0] == "RNaseH":
                    RNaseH_list.append(each_hit)

                elif match_domain.split("_")[0] == "INT":
                    INT_list.append(each_hit)

                elif match_domain.split("_")[0] == "ENV":
                    ENV_list.append(each_hit)


            hit_GAG = best_domain(GAG_list, 1)

            hit_DUT = best_domain(DUT_list, 1)

            hit_AP = best_domain(AP_list, 1)

            hit_RT = best_domain(RT_lsit, 1)

            hit_RNaseH = best_domain(RNaseH_list, 1)

            hit_INT = best_domain(INT_list, 1)

            hit_ENV = best_domain(ENV_list, 1)


            if hit_GAG != "not found":
                if int(hit_GAG.split(",")[-1]) >= self.para_dic["GAG_length"]:
                    line_flage = "True"

                    classify = hit_GAG.split(",")[0].split("_")[-1]

                    if classify not in retroviridae_list:
                        retroviridae_flage = False

                else:
                    hit_GAG = "relatively short"



            if hit_DUT != "not found":
                line_flage = "True"


            if hit_AP != "not found":
                if int(hit_AP.split(",")[-1]) >= self.para_dic["PRO_length"]:
                    line_flage = "True"

                else:
                    hit_AP = "relatively short"


            if hit_RT != "not found":
                if int(hit_RT.split(",")[-1]) >= self.para_dic["RT_length"]:
                    line_flage = "True"

                    classify = hit_RT.split(",")[0].split("_")[-1]

                    if classify not in retroviridae_list:
                        retroviridae_flage = False

                else:
                    hit_RT = "relatively short"


            if hit_RNaseH != "not found":
                if int(hit_RNaseH.split(",")[-1]) >= self.para_dic["RNaseH_length"]:
                    line_flage = "True"

                else:
                    hit_RNaseH = "relatively short"


            if hit_INT != "not found":
                if int(hit_INT.split(",")[-1]) >= self.para_dic["INT_length"]:
                    line_flage = "True"

                else:
                    hit_INT = "relatively short"


            if hit_ENV != "not found":
                if int(hit_ENV.split(",")[-1]) >= self.para_dic["ENV_length"]:
                    line_flage = "True"

                else:
                    hit_ENV = "relatively short"


            if line_flage == "True" and retroviridae_flage == True:
                HMMER_simplify_file.write(
                    key + "\t" + hit_GAG + "\t" + hit_DUT + "\t" + hit_AP
                    + "\t" + hit_RT + "\t" + hit_RNaseH + "\t" + hit_INT
                    + "\t" + hit_ENV + "\n")

        hmmer_result_file.close()
        HMMER_simplify_file.close()



# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/10 18:40

"""
import sys
import os
import platform

def reversecomp_seq(input_str):
    """
    Reverse complement of sequence
    :param input_str: str
    :return: str
    """

    trantab1 = str.maketrans('ACGTRYMKVBHDSWNacgtrymkvbhdswn',
                             'TGCAYRKMBVDHWSNtgcayrkmbvdhwsn')

    trantab2 = str.maketrans('ACGURYMKVBHDSWNacgurymkvbhdswn',
                             'UGCAYRKMBVDHWSNugcayrkmbvdhwsn')

    line = input_str.strip()

    line = line[::-1]  # reverse sequence

    if line.find("T") != -1 or line.find("t") != -1:
        line = line.translate(trantab1)
    else:
        line = line.translate(trantab2)

    return line



def make_dir(input_dir):
    """
    :param input_dir: directory which need to create
    :return:
    """
    if os.path.isdir(input_dir) == False:
        os.makedirs(input_dir)


def get_all_path(open_dir_path):
    """
    Gets all the path of files in a folder
    :param open_dir_path: folder
    :return: path of files
    """

    rootdir = open_dir_path

    path_list = []

    lists = os.listdir(rootdir)

    for i in range(0, len(lists)):
        com_path = os.path.join(rootdir, lists[i])

        if platform.system().lower() == "windows":
            com_path = com_path.replace("\\", "/")
        # print(com_path)

        if os.path.isfile(com_path):
            path_list.append(com_path)


    return path_list



def read_gff3(input_path):
    """
    :param input: gff3 file
    :return: seqs and annotation dic
    """
    gff3_file = open(input_path,"r",encoding="utf-8")
    line_num = 0
    seq_nums = []
    i = 0
    seqs = {}  # sequence name

    annotations = {}

    for line in gff3_file:
        line_num += 1
        line = line.strip()
        
        if line_num >= 2 and line != "":
            if line.startswith("##seq"):
                seq_num = line.split(" ")[3]
                # print(str(seq_num))
                seq_nums.append(seq_num)

            elif line.startswith("###"):
                pass

            elif line.startswith("#"):
                seqs[seq_nums[i]] = line[1:]
                i += 1

            else:
                anno_list = line.split("\t")

                seq_name = anno_list[0]

                if seq_name not in annotations:
                    annotations[seq_name] = [line]

                else:
                    annotations[seq_name].append(line)



    gff3_file.close()

    return seqs, annotations



def best_domain(input_list,n):
    """
    After sorting the nested list, a list with the smallest element in it
    :param input_list: Enter a nested list
    :param n: number of column used for sort
    :return: the best domain
    """
    hit_domain = "not found"

    if len(input_list) >= 1:
        hit_list_sort = sorted(input_list, key=lambda x: x[n])

        hit_list = hit_list_sort[0]

        hit_domain = (hit_list[0] + "," + str(hit_list[1]) + "," + hit_list[2]
                            + "," + str(hit_list[3]) + "," + str(hit_list[4]))
        # name of domain, c-Evalue, ORF, Start and end point, lenth of domain


    return hit_domain


def domain_locate_genome(seq_coordinates, seq_stand, domain_aa_range,orf):
    """
    domain relocation on the original genome

    :param seq_coordinates: The coordinates of the sequence on the original
    genome, for example: "2000..4902"

    :param seq_stand: The direction of the sequence, "+" means the same as the
    input genome, "-" means the complementary strand

    :param domain_coordinates: The coordinates (amino acids) of the domain
    predicted by HMMER in the sequence, the format is for example: "100..202"

    :param orf: When the domain is predicted, which set of coding frames of
    the sequence is used, such as "ORF1"

    :return: The location on the original genome (for example: "1900..2202")
    and the chain where the domain is located
    """
    seq_coordinates_start = int(seq_coordinates.split("..")[0])
    seq_coordinates_end = int(seq_coordinates.split("..")[1])

    seq_lenth = seq_coordinates_end - seq_coordinates_start + 1

    domain_aa_range = (domain_aa_range.split("(")[-1]).split(")")[0]

    domain_in_strand = "ERROR"
    site_range = "(..)"


    if seq_stand == "same":
        if orf == "ORF1":
            domain_nt_start = 3 *(int(domain_aa_range.split("..")[0]) - 1)  + 1
            domain_nt_end = 3 * int(domain_aa_range.split("..")[1])

            genome_start = seq_coordinates_start + domain_nt_start - 1
            genome_end = seq_coordinates_start + domain_nt_end - 1

            domain_in_strand = "same"
            site_range = "(" + str(genome_start) + ".." + str(genome_end) + ")"

        elif orf == "ORF2":
            domain_nt_start = 3 * (int(domain_aa_range.split("..")[0]) - 1) + 1 + 1
            domain_nt_end = 3 * int(domain_aa_range.split("..")[1]) + 1
            genome_start = seq_coordinates_start + domain_nt_start - 1
            genome_end = seq_coordinates_start + domain_nt_end - 1

            domain_in_strand = "same"
            site_range = "(" + str(genome_start) + ".." + str(genome_end) + ")"

        elif orf == "ORF3":
            domain_nt_start = 3 * (
                        int(domain_aa_range.split("..")[0]) - 1) + 1 + 2
            domain_nt_end = 3 * int(domain_aa_range.split("..")[1]) + 2
            genome_start = seq_coordinates_start + domain_nt_start - 1
            genome_end = seq_coordinates_start + domain_nt_end - 1

            domain_in_strand = "same"
            site_range = "(" + str(genome_start) + ".." + str(genome_end) + ")"


    elif seq_stand == "complement":
        if orf == "ORF1":
            domain_nt_start = 3 * (
                    int(domain_aa_range.split("..")[0]) - 1) + 1
            domain_nt_end = 3 * int(domain_aa_range.split("..")[1])

            domain_nt_restart = seq_lenth - domain_nt_end + 1
            domain_nt_reend = seq_lenth - domain_nt_start + 1

            genome_start = seq_coordinates_start + domain_nt_restart - 1
            genome_end = seq_coordinates_start + domain_nt_reend - 1

            domain_in_strand = "complement"
            site_range = ("complement(" + str(genome_start)
                          + ".." + str(genome_end) + ")")

        elif orf == "ORF2":
            domain_nt_start = 3 * (
                    int(domain_aa_range.split("..")[0]) - 1) + 1 + 1
            domain_nt_end = 3 * int(domain_aa_range.split("..")[1]) + 1

            domain_nt_restart = seq_lenth - domain_nt_end + 1
            domain_nt_reend = seq_lenth - domain_nt_start + 1

            genome_start = seq_coordinates_start + domain_nt_restart - 1
            genome_end = seq_coordinates_start + domain_nt_reend - 1

            domain_in_strand = "complement"
            site_range = ("complement(" + str(genome_start)
                         + ".." + str(genome_end) + ")")

        elif orf == "ORF3":
            domain_nt_start = 3 * (
                    int(domain_aa_range.split("..")[0]) - 1) + 1 + 2
            domain_nt_end = 3 * int(domain_aa_range.split("..")[1]) + 2

            domain_nt_restart = seq_lenth - domain_nt_end + 1
            domain_nt_reend = seq_lenth - domain_nt_start + 1

            genome_start = seq_coordinates_start + domain_nt_restart - 1
            genome_end = seq_coordinates_start + domain_nt_reend - 1

            domain_in_strand = "complement"
            site_range = ("complement(" + str(genome_start)
                          + ".." + str(genome_end) + ")")



    return (domain_in_strand,site_range)



def hmmer_contain(input_path):

    domain = []
    line_num = 0
    with open(input_path,"r", encoding="utf-8") as input_file:
        for line in input_file:
            line_num += 1
            if line_num <= 5:
                if not line.startswith("#"):
                    domain.append(line.strip())

    if domain == []:
        return "False"

    else:
        return "True"





# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/11 22:22

"""

class PairLTRErvAnno(object):
    def __init__(self, pairltr_ervs_map_genome, pairltr_ervs_anno, out_refix):

        """
        annotate domain  on the extracted pair of LTR_ervs
        :param pairltr_ervs_map_genome:
        :param pairltr_ervs_anno:
        :param out_refix:
        """
        super(PairLTRErvAnno,self).__init__()

        self.pairltr_ervs_map_genome = pairltr_ervs_map_genome
        self.pairltr_ervs_anno = pairltr_ervs_anno
        self.out_refix = out_refix
    
    def run(self):            

        domain_in_genome_file = open(self.pairltr_ervs_map_genome, "r",
                                     encoding="utf-8")
       
        out_domain_annos = open(self.pairltr_ervs_anno,"w",encoding="utf-8")

        out_domain_annos.write("##Name of ERVs with paired-LTRs" + "\t"
                               + "5'-LTR region" + "\t"
                               + "3'-LTR region" + "\t"
                               + "GAG" + "\t"
                               + "DUT" + "\t"
                               + "AP" + "\t"
                               + "RT" + "\t"
                               + "RNaseH" + "\t"
                               + "INT"+ "\t"
                               + "ENV" + "\n")


        for line in domain_in_genome_file:

            if not line.startswith("##"):

                line = line.strip()
                line_list = line.split("\t")

                seq_name = line_list[0]
                strand = seq_name.split("|")[-1]

                flage = 0

                line_record_list = []

                if strand == "same":
                    start_site = int(seq_name.split("|")[1])
                    # end_site =int(seq_name.split("|")[2])
                    for n in range(1, len(line_list)):
                        if n == 1 or n == 2:
                            sites = line_list[n]

                            LTR_start = int(
                                sites.split("..")[0]) - start_site + 1

                            LTR_end = int(sites.split("..")[1]) - start_site + 1

                            line_record_list.append(str(LTR_start) + ".."
                                                    + str(LTR_end))

                        elif n == 3:
                            pass

                        else:
                            if line_list[n] != "not found" and line_list[n].find("relatively short") == -1:
                                xx = (line_list[n].split(",")[-1]).split("(")[0]
                                if xx == "":
                                    sites = \
                                    (line_list[n].split("(")[1]).split(")")[0]
                                    domain_start = int(
                                        sites.split("..")[0]) - start_site + 1
                                    domain_end = int(
                                        sites.split("..")[-1]) - start_site + 1
                                    line_record_list.append(
                                        str(domain_start) + ".."
                                        + str(domain_end))
                                elif xx == "complement":
                                    sites = \
                                    (line_list[n].split("(")[1]).split(")")[0]
                                    domain_start = int(
                                        sites.split("..")[0]) - start_site + 1
                                    domain_end = int(
                                        sites.split("..")[-1]) - start_site + 1
                                    line_record_list.append("complement" + "("
                                                            + str(
                                        domain_start) + ".."
                                                            + str(
                                        domain_end) + ")")
                                    flage = 1

                            else:
                                line_record_list.append(line_list[n])


                elif strand == "complement":
                    start_site = int(seq_name.split("|")[2])
                    for n in range(1, len(line_list)):

                        if n == 1 or n == 2:

                            ltr_sites = (line_list[n].split("(")[1]).split(")")[
                                0]

                            LTR_start = start_site - int(
                                ltr_sites.split("..")[1]) + 1

                            LTR_end = start_site - int(
                                ltr_sites.split("..")[0]) + 1

                            line_record_list.append(str(LTR_start) + ".."
                                                    + str(LTR_end))
                        elif n == 3:
                            pass

                        else:
                            if line_list[n] != "not found" and line_list[n].find("relatively short") == -1:
                                xx = (line_list[n].split(",")[-1]).split("(")[0]
                                if xx == "complement":
                                    sites = \
                                    (line_list[n].split("(")[1]).split(")")[0]
                                    domain_start = start_site - int(
                                        sites.split("..")[1]) + 1
                                    domain_end = start_site - int(
                                        sites.split("..")[0]) + 1

                                    line_record_list.append(
                                        str(domain_start) + ".."
                                        + str(domain_end))
                                elif xx == "":
                                    sites = \
                                    (line_list[n].split("(")[1]).split(")")[0]
                                    domain_start = start_site - int(
                                        sites.split("..")[1]) + 1
                                    domain_end = start_site - int(
                                        sites.split("..")[0]) + 1

                                    line_record_list.append("complement" + "("
                                                            + str(
                                        domain_start) + ".."
                                                            + str(
                                        domain_end) + ")")
                                    flage = 1

                            else:
                                line_record_list.append(line_list[n])

                new_line = "\t".join(line_record_list)

                out_domain_annos.write(seq_name + "\t"
                    + new_line + "\n")

                if flage == 1:
                    out_domain_annos.write(
                        "##Warn: the annotation and chain direction of '" +
                        seq_name
                        + "' needs to be checked in file of 'domain_annotation_in_genome_final.txt'!" + "\n")


        domain_in_genome_file.close()
        out_domain_annos.close()



class PotentialEveAnno(object):
    def __init__(self, map_duplicate_removal,
                 domain_in_potential_erv,
                 out_refix):
        self.out_map_duplicate_removal = map_duplicate_removal
        self.domain_in_potential_erv = domain_in_potential_erv
        self.out_refix = out_refix

        super(PotentialEveAnno, self).__init__()

    def run(self):

        domain_duplicate_removal = open(self.out_map_duplicate_removal, "r",
                                        encoding="utf-8")
        out_domain_annos = open(self.domain_in_potential_erv, "w",
                                encoding="utf-8")

        out_domain_annos.write("##Name of potential ERVLEs without paired-LTRs" + "\t"
                               + "GAG" + "\t"
                               + "DUT" + "\t"
                               + "AP" + "\t"
                               + "RT" + "\t"
                               + "RNaseH" + "\t"
                               + "INT" + "\t"
                               + "ENV" + "\n")


        for line in domain_duplicate_removal:

            if not line.startswith("##"):

                line = line.strip()
                line_list = line.split("\t")

                seq_name = line_list[0]
                strand = seq_name.split("|")[-1]

                flage = 0

                line_record_list = []
                if strand == "same":
                    start_site = int(seq_name.split("|")[1])
                    # end_site =int(seq_name.split("|")[2])
                    for n in range(1, len(line_list)):
                        if line_list[n] != "not found" and line_list[n].find("relatively short") == -1:
                            xx = (line_list[n].split(",")[-1]).split("(")[0]
                            if xx == "":
                                sites = (line_list[n].split("(")[1]).split(")")[
                                    0]
                                domain_start = int(
                                    sites.split("..")[0]) - start_site + 1
                                domain_end = int(
                                    sites.split("..")[-1]) - start_site + 1
                                line_record_list.append(str(domain_start) + ".."
                                                        + str(domain_end))
                            elif xx == "complement":
                                sites = (line_list[n].split("(")[1]).split(")")[
                                    0]
                                domain_start = int(
                                    sites.split("..")[0]) - start_site + 1
                                domain_end = int(
                                    sites.split("..")[-1]) - start_site + 1
                                line_record_list.append("complement" + "("
                                                        + str(
                                    domain_start) + ".."
                                                        + str(domain_end) + ")")
                                flage = 1
                        else:
                            line_record_list.append(line_list[n])


                elif strand == "complement":
                    start_site = int(seq_name.split("|")[2])
                    for n in range(1, len(line_list)):

                        if line_list[n] != "not found" and line_list[n].find("relatively short") == -1:

                            xx = (line_list[n].split(",")[-1]).split("(")[0]
                            if xx == "complement":
                                sites = (line_list[n].split("(")[1]).split(")")[
                                    0]
                                domain_start = start_site - int(
                                    sites.split("..")[1]) + 1
                                domain_end = start_site - int(
                                    sites.split("..")[0]) + 1

                                line_record_list.append(str(domain_start) + ".."
                                                        + str(domain_end))
                            elif xx == "":
                                sites = (line_list[n].split("(")[1]).split(")")[
                                    0]
                                domain_start = start_site - int(
                                    sites.split("..")[1]) + 1
                                domain_end = start_site - int(
                                    sites.split("..")[0]) + 1

                                line_record_list.append("complement" + "("
                                                        + str(
                                    domain_start) + ".."
                                                        + str(domain_end) + ")")
                                flage = 1

                        else:
                            line_record_list.append(line_list[n])

                new_line = "\t".join(line_record_list)

                out_domain_annos.write(seq_name + "\t"
                    + new_line + "\n")

                if flage == 1:
                    out_domain_annos.write(
                        "##Warn: the annotation and chain direction of '" +
                        seq_name
                        + "' needs to be checked in file of 'domain_annotation_in_genome_final.txt'!" + "\n")


        domain_duplicate_removal.close()
        out_domain_annos.close()





# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/11 17:47

"""
from Bio import SeqIO
from public_function import reversecomp_seq


class GetPairLTRervs(object):
    def __init__(self, domain_map_genome, host_path, doamin_seq_dir,
                  erv_seq_path, pairLTRs_path):
        """
        Extract pairLTR ERVs and domains
        :param domain_map_genome:
        :param host_path:
        :param doamin_seq_dir:
        :param erv_seq_path:
        :param pairLTRs_path:
        """
        super(GetPairLTRervs,self).__init__()

        self.domain_map_genome = domain_map_genome
        self.doamin_seq_dir = doamin_seq_dir
        self.host_path = host_path
        self.erv_seq_path = erv_seq_path
        self.pairLTRs_path = pairLTRs_path
    
    def run(self):

        domain_in_genome_file = open(self.domain_map_genome,"r",encoding="utf-8")

        host_seq_dict = SeqIO.to_dict(SeqIO.parse(self.host_path, "fasta")) # 宿主基因组

        erv_seq = open(self.erv_seq_path,"w",encoding="utf-8")

        out_pair_LTRs = open(self.pairLTRs_path,"w",encoding="utf-8")
        
        for line in domain_in_genome_file:
            if not line.startswith("##"):
        
                line_list = line.split("\t")
                seq_name = line_list[0]
                chain_strand = seq_name.split("|")[-1]
        
                erv_start = int(seq_name.split("|")[1])
                erv_end = int(seq_name.split("|")[2])
        
                key = seq_name.split("|")[0]
        
                five_ltr = line_list[1]
                three_ltr = line_list[2]
        
                if chain_strand == "same":
        
                    ervs_seq_contain = str(host_seq_dict[key].seq[erv_start-1 :erv_end])
        
                    five_ltr_start= int(five_ltr.split("..")[0])
                    five_ltr_end = int(five_ltr.split("..")[1])
                    five_ltr_seq = str(host_seq_dict[key].seq[five_ltr_start - 1: five_ltr_end])
        
                    three_ltr_start = int(three_ltr.split("..")[0])
                    three_ltr_end = int(three_ltr.split("..")[1])
                    three_ltr_seq = str(host_seq_dict[key].seq[three_ltr_start - 1: three_ltr_end])
        
        
        
                else:
                    ervs_seq_contain = reversecomp_seq(str(host_seq_dict[key].seq[erv_start - 1:erv_end]))
        
                    five_ltr_site = (five_ltr.split("(")[-1]).split(")")[0]
                    five_ltr_start = int(five_ltr_site.split("..")[0])
                    five_ltr_end = int(five_ltr_site.split("..")[1])
                    five_ltr_seq = reversecomp_seq(str(
                        host_seq_dict[key].seq[five_ltr_start - 1: five_ltr_end]))
        
                    three_ltr_site = (three_ltr.split("(")[-1]).split(")")[0]
                    three_ltr_start = int(three_ltr_site.split("..")[0])
                    three_ltr_end = int(three_ltr_site.split("..")[1])
                    three_ltr_seq = reversecomp_seq(str(
                        host_seq_dict[key].seq[three_ltr_start - 1: three_ltr_end]))
        
        
                erv_seq.write(">" + seq_name + "\n" + ervs_seq_contain + "\n")
                out_pair_LTRs.write(">" + seq_name + "_five_ltr" + "\n"
                                    + five_ltr_seq + "\n"
                                    + ">" + seq_name + "_three_ltr" + "\n"
                                    + three_ltr_seq + "\n")


                for n in range(4,len(line_list)):
                    domain_record = line_list[n].strip()
                    if domain_record != "not found" and domain_record.find("relatively short") == -1:
                        domain_name = (domain_record.split(",")[0]).split("_")[0]
        
                        domain_range = (domain_record.split("(")[-1]).strip(")")
        
                        domain_start = int(domain_range.split("..")[0])
                        domain_end = int(domain_range.split("..")[1])
        
                        if (domain_record.split(",")[-1]).split("(")[0]=="complement":
                            domain_seq = reversecomp_seq(str(host_seq_dict[key].seq[domain_start-1 :domain_end]))
        
        
                        else:
                            domain_seq = str(host_seq_dict[key].seq[domain_start-1 :domain_end])
        
        
                        out_domain = open(self.doamin_seq_dir + r"/" + domain_name + ".fasta", "a",
                            encoding="utf-8")
        
                        out_domain.write(">" + seq_name + "_" + domain_name + "\n"
                              +  domain_seq + "\n")

        domain_in_genome_file.close()
        erv_seq.close()
        out_pair_LTRs.close()


class GetPotentialEves(object):
    def __init__(self, map_duplicate_removal, host_path,
                 doamin_seq_dir, erv_seq_path):
        """
        Extract potential ERVs and domains on the flank sequence
        :param map_duplicate_removal:
        :param host_path:
        :param doamin_seq_dir:
        :param erv_seq_path:
        """
        super(GetPotentialEves, self).__init__()

        self.map_duplicate_removal = map_duplicate_removal
        self.host_path = host_path
        self.doamin_seq_dir = doamin_seq_dir
        self.erv_seq_path = erv_seq_path

    def run(self):
        domain_duplicate_removal = open(self.map_duplicate_removal, "r",
                                        encoding="utf-8")

        host_seq_dict = SeqIO.to_dict(SeqIO.parse(self.host_path, "fasta"))

        erv_seq = open(self.erv_seq_path, "w",encoding="utf-8")

        for line in domain_duplicate_removal:
            if not line.startswith("##"):
                line_list = line.split("\t")
                seq_name = line_list[0]
                chain_strand = seq_name.split("|")[-1]

                erv_start = int(seq_name.split("|")[1])
                erv_end = int(seq_name.split("|")[2])

                key = seq_name.split("|")[0]

                if chain_strand == "same":

                    ervs_seq_contain = str(
                        host_seq_dict[key].seq[erv_start - 1:erv_end])

                else:
                    ervs_seq_contain = reversecomp_seq(
                        str(host_seq_dict[key].seq[erv_start - 1:erv_end]))

                erv_seq.write(">" + seq_name + "\n" + ervs_seq_contain + "\n")


                for n in range(1, len(line_list)):
                    domain_record = line_list[n].strip()
                    if domain_record != "not found" and domain_record.find("relatively short") == -1:

                        domain_name = (domain_record.split(",")[0]).split("_")[
                            0]

                        domain_range = (domain_record.split("(")[-1]).strip(")")

                        domain_start = int(domain_range.split("..")[0])
                        domain_end = int(domain_range.split("..")[1])

                        if (domain_record.split(",")[-1]).split("(")[
                            0] == "complement":
                            domain_seq = reversecomp_seq(str(
                                host_seq_dict[key].seq[
                                domain_start - 1:domain_end]))
                        else:
                            domain_seq = str(host_seq_dict[key].seq[
                                             domain_start - 1:domain_end])

                        out_domain = open(self.doamin_seq_dir + r"/"
                                          + domain_name + ".fasta", "a",
                                          encoding="utf-8")

                        out_domain.write(
                            ">" + seq_name + "_" + domain_name + "\n"
                            + domain_seq + "\n")

        domain_duplicate_removal.close()
        erv_seq.close()





# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/9 15:19

"""

import sys
import os
import platform
# import psutil

app_dir = os.path.dirname(os.path.realpath(__file__))
# print(app_dir)


if platform.system().lower() == "windows":
    app_dir = app_dir.replace("\\", "/")

# sys.path.append(app_dir) # has been in __init__


import time
import subprocess
import zipfile
import argparse

from datetime import datetime
from get_flank import GetFlanks
from simplify_gff3 import SimplyGFF
from seqs_for_hmmer import GetSeqForHmmer,GetSeqForHmmer2
from simplify_hmmer import SimplyHMMER
from map_genome import PairLTRervMap, NoPairLTReveMap
from get_ervs import GetPairLTRervs, GetPotentialEves
from annotion_ervs import PairLTRErvAnno, PotentialEveAnno
from public_function import make_dir,hmmer_contain,get_all_path



#  ***** External program *****


makeblastdb_execute = "makeblastdb"     # from blast
tblastn_execute = "tblastn"             # from blast
gt_execute = "gt"                       # from genometools
hmmpress_execute = "hmmpress"           # from hmmer
hmmscan_execute = "hmmscan"             # from hmmer



if platform.system().lower() == "windows":
    makeblastdb_execute = "makeblastdb.exe"
    tblastn_execute = "tblastn.exe"
    gt_execute = "gt.exe"
    hmmpress_execute = "hmmpress.exe"
    hmmscan_execute = "hmmscan.exe"




#  ***** Built-in database *****

ERVprofiles_zip = app_dir + "/Internal_database/ERVprofiles/ERVprofile.zip"
ERVprofiles_path = app_dir + "/Internal_database/ERVprofiles/ERV.profile"
probe_dir =  app_dir + "/Internal_database/probe"



example_use = r'''
----------------☆ Example of use ☆-----------------

ervdetective -i myotis_lucifugus.fna -p myotis_lucifugus -n 10 -o output
  
----------------------☆  End  ☆---------------------

'''

"""
def calculate_memory():
    pid = os.getpid()
    p = psutil.Process(pid)
    info = p.memory_full_info()
    memory = info.uss / 1024 / 1024
    return memory
"""


# ###  *** Step 0, Processing input parameters  **

def parameter():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        prog="ervdetective",
        description="",
        epilog=example_use)

    parser.add_argument(
        "-i", dest="host",
        help="The file-path of host genome sequence, the suffix is generally *.fna, *.fas, *.fasta.",
        default="")

    parser.add_argument(
        "-n", dest="thread",
        help="Specify the number of threads used, default: 1.",
        type=int,
        default=1)

    parser.add_argument(
        "-p", dest="prefix",
        help="The prefix of output file, default character: 'host'.",
        type=str,
        default="host")

    parser.add_argument(
        "-o", dest="output",
        help="The path of output folder to store all the results.",
        default="")

    parser.add_argument(
        "-eb", dest="eblast",
        help="Specify threshold of e-value for BLAST search, default: 1e-5.",
        type=float,
        default=0.00001)

    parser.add_argument(
        "-f", dest="flank",
        help="The length of extended flank sequence on either side of "
             "the blast hit-site, default: 15000.",
        type=int,
        default=15000)

    parser.add_argument(
        "-l1", dest="minltr",
        help="Specify minimum length of LTR, default: 100.",
        type=int,
        default=100)

    parser.add_argument(
        "-l2", dest="maxltr",
        help="Specify maximum length of LTR, default: 1000.",
        type=int,
        default=1000)

    parser.add_argument(
        "-s", dest="ltrsimilar",
        help="Specify threshold(%%) of the similarity of paired LTRs, default: 80.",
        type=float,
        default=80)

    parser.add_argument(
        "-d1", dest="mindistltr",
        help="The minimum interval of paired-LTRs start-positions, default: 1000.",
        type=int,
        default=1000)

    parser.add_argument(
        "-d2", dest="maxdistltr",
        help="The maximum interval of paired-LTRs start-positions, default: 15000.",
        type=int,
        default=15000)

    parser.add_argument(
        "-t1", dest="mintsd",
        help="The minimum length for each TSD site, default: 4.",
        type=int,
        default=4)

    parser.add_argument(
        "-t2", dest="maxtsd",
        help="The maximum length for each TSD site, default: 6.",
        type=int,
        default=6)

    parser.add_argument(
        "-motif", dest="motif",
        help="Specify start-motif (2 nucleotides) and end-motif (2 nucleotides), default string: TGCA.",
        type=str,
        default="TGCA")

    parser.add_argument(
        "-mis", dest="mismotif",
        help="The maximum number of mismatches nucleotides in motif, default: 1.",
        type=int,
        default=1)

    parser.add_argument(
        "-ed", dest="ehmmer",
        help="Specify threshold of e-value using for HMMER search, default: 1e-6.",
        type=float,
        default=0.000001)

    parser.add_argument(
        "--gag", dest="GAG_length",
        help="The threshold of length of GAG protein in HMMER search, default: 250 aa.",
        type=int,
        default=250)

    parser.add_argument(
        "--pro", dest="PRO_length",
        help="The threshold of length of PRO protein in HMMER search, default: 50 aa.",
        type=int,
        default=50)

    parser.add_argument(
        "--rt", dest="RT_length",
        help="The threshold of length of RT protein in HMMER search, default: 150 aa.",
        type=int,
        default=150)

    parser.add_argument(
        "--rh", dest="RNaseH_length",
        help="The threshold of length of RNaseH protein in HMMER search, default: 65 aa.",
        type=int,
        default=65)

    parser.add_argument(
        "--int", dest="INT_length",
        help="The threshold of length of INT protein in HMMER search, default: 150 aa.",
        type=int,
        default=150)

    parser.add_argument(
        "--env", dest="ENV_length",
        help="The threshold of length of ENV protein in HMMER search, default: 250 aa.",
        type=int,
        default=250)



    theargs = parser.parse_args(sys.argv[1:])

    return theargs


def starts():

    print("\n" + "-------------------------------------------------")

    print("  Name: Endogenous retroviruses detective (ERVdetective)")

    print("  Description: An efficient pipeline for identification and annotation of endogenous retroviruses.")

    print("  Version: 1.0.9 (2024-05-01)")

    print("  Author: Zhi-Jian Zhou")

    print("-------------------------------------------------" + "\n")


    myargs = parameter()
    
    print(myargs)
    
    print("\n")

    time_start = datetime.today().now()

    # start_memory = calculate_memory()

    parameter_dic = {}

    parameter_dic["host_path"] = myargs.host

    parameter_dic["out_prefix"] = myargs.prefix

    parameter_dic["out_dir"] = myargs.output

    parameter_dic["thread"] = str(myargs.thread)

    parameter_dic["blastE"] = str(myargs.eblast)

    parameter_dic["flanklen"] = myargs.flank

    parameter_dic["minlength"] = str(myargs.minltr)

    parameter_dic["maxlength"] = str(myargs.maxltr)

    parameter_dic["LTRsimilar"] = str(myargs.ltrsimilar)

    parameter_dic["mindistltr"] = str(myargs.mindistltr)

    parameter_dic["maxdistltr"] = str(myargs.maxdistltr)

    parameter_dic["mintsd"] = str(myargs.mintsd)

    parameter_dic["maxtsd"] = str(myargs.maxtsd)

    parameter_dic["motif_str"] = myargs.motif

    parameter_dic["motifmis"] = str(myargs.mismotif)

    parameter_dic["HmmerE"] = str(myargs.ehmmer)

    parameter_dic["GAG_length"] = int(myargs.GAG_length)

    parameter_dic["PRO_length"] = int(myargs.PRO_length)

    parameter_dic["RT_length"] = int(myargs.RT_length)

    parameter_dic["RNaseH_length"] = int(myargs.RNaseH_length)

    parameter_dic["INT_length"] = int(myargs.INT_length)

    parameter_dic["ENV_length"] = int(myargs.ENV_length)



    if platform.system().lower() == "windows":
        parameter_dic["host_path"] = parameter_dic["host_path"].replace("\\", "/")
        parameter_dic["out_dir"] = parameter_dic["out_dir"].replace("\\", "/")


    run_log_dir = parameter_dic["out_dir"] + "/Run_log"

    make_dir(parameter_dic["out_dir"])
    make_dir(run_log_dir)


    # When the input and output are not right
    if parameter_dic["host_path"] == "":
        print("Error: the path of input host-genome was not specified! Please use "
              "'ervdetective -h' to view help document.")
        sys.exit()

    if parameter_dic["out_dir"] == "":
        print("Error: the path of out folder was not specified! Please use "
              "'ervdetective -h' to view help document.")
        sys.exit()


    ###  *** Step 1, run tblastn ***

    print(">>> Blast search for ERVs loci....")

    blast_result_path = run_log_dir + "/Record1_blast_result"

    host_db_dir = blast_result_path + "/seqdb"

    blast_out_dir = blast_result_path + "/blast_result"

    make_dir(blast_result_path)

    make_dir(host_db_dir)

    make_dir(blast_out_dir)


    makeblastdb_command = (makeblastdb_execute + " -in "
                            + parameter_dic["host_path"]
                            + " -dbtype nucl "
                            + "-out " + host_db_dir + "/"
                            + parameter_dic["out_prefix"]
                            + ".")


    makeblastdb_process = subprocess.Popen(makeblastdb_command,
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.STDOUT,
                                           universal_newlines=True,
                                           shell=True)

    while True:

        output = makeblastdb_process.stdout.readline()
        if makeblastdb_process.poll() is not None:
            break

        elif output:
            print(output)

    makeblastdb_process.terminate()


    # blast process
    probe_file_list = get_all_path(probe_dir)

    tblastn_commd_list = []
    for each_path in probe_file_list:
        each_path = each_path.strip()
        file_name = str(each_path.split("/")[-1]).strip()
        print("\n" + ">>> Use the " +  file_name + " as a probe...")
        blast_result = (blast_out_dir + "/"
                        + parameter_dic["out_prefix"]
                        + "_" + file_name + ".txt")

        each_tblastn_commd = (tblastn_execute + " -query " + each_path
                              + " -db " + host_db_dir + "/"
                              + parameter_dic["out_prefix"] + "."
                              + " -out " + blast_result
                              + " -outfmt 6"
                              + " -evalue " + parameter_dic["blastE"]
                              + " -num_threads " + parameter_dic["thread"])

        tblastn_commd_list.append(each_tblastn_commd)


    tblastn_commd = " && ".join(tblastn_commd_list)

    tblastn_process = subprocess.Popen(tblastn_commd,
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.STDOUT,
                                           universal_newlines=True,
                                           shell=True)

    while True:

        output = tblastn_process.stdout.readline()
        if tblastn_process.poll() is not None:
            break

        elif output:
            print(output)

    tblastn_process.terminate()


    # Determine whether the blast result file is empty

    blast_file_list = get_all_path(blast_out_dir)

    blast_file_flage = "False"
    for each_file in blast_file_list:
        file_size = os.path.getsize(each_file)
        if file_size > 0:
            blast_file_flage = "True"


    if blast_file_flage == "False":
        print("\n"
              + "*** Note: no homologues of ERVs were found in the input sequence!"
              + "\n")

        sys.exit()


    ###  ***Step 2, get flank sequence on both sides of these hit sites***

    flank_seq_dir = run_log_dir + "/Record2_flank_seqs"

    make_dir(flank_seq_dir)

    flank_seq_file = flank_seq_dir + "/blast_flank_seq.fasta"

    flank_get = GetFlanks(parameter_dic["host_path"], blast_out_dir,
                          parameter_dic["flanklen"], flank_seq_file)
    flank_get.run()



    ###  ***Step 3, identify LTR by LTRharvest***

    print("\n" + ">>> Search pair-LTR....")

    identy_pairLTR = run_log_dir + "/Record3_identy_pairedLTR"

    flankseq_index_dir = identy_pairLTR + "/flank_seq_index"

    ltrharvest_result_dir = identy_pairLTR + "/LTRharvest_result"

    flankseq_index_file = flankseq_index_dir + "/FlankSeq"

    ltrharvest_gff3_file = ltrharvest_result_dir + "/flank_seq_harvest.gff3"
    ltrharvest_out_seq = ltrharvest_result_dir + "/LTRharvest_out.fas"
    harvest_log_file = ltrharvest_result_dir + "/harvest_log.txt"

    make_dir(identy_pairLTR)
    make_dir(flankseq_index_dir)
    make_dir(ltrharvest_result_dir)

    make_index_commd = (gt_execute + " suffixerator"
                          + " -db " + flank_seq_file
                          + " -indexname " + flankseq_index_file
                          + " -tis -suf -lcp -des -ssp -sds -dna")

    ltrharvest_commd = (gt_execute + " ltrharvest"
                      + " -index " + flankseq_index_file
                      + " -minlenltr " + parameter_dic["minlength"]
                      + " -maxlenltr " + parameter_dic["maxlength"]
                      + " -similar " + parameter_dic["LTRsimilar"]
                      + " -seed 20"
                      + " -mindistltr " + parameter_dic["mindistltr"]
                      + " -maxdistltr " + parameter_dic["maxdistltr"]
                      + " -mintsd " + parameter_dic["mintsd"]
                      + " -maxtsd " + parameter_dic["maxtsd"]
                      + " -motif " + parameter_dic["motif_str"]
                      + " -motifmis " + parameter_dic["motifmis"]
                      + " -gff3 " + ltrharvest_gff3_file
                      + " -out " + ltrharvest_out_seq
                      + " > " + harvest_log_file)



    find_ltr_commd = make_index_commd + " && " + ltrharvest_commd


    find_ltr_process = subprocess.Popen(find_ltr_commd, stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               universal_newlines=True, shell=True)

    while True:

        output = find_ltr_process.stdout.readline()
        if find_ltr_process.poll() is not None:
            break

        elif output:
            print(output)

    find_ltr_process.terminate()



    ###  ***Step4, Prepare the sequence for HMMER ***

    print("\n" + ">>> Prepare the sequence for HMMER...")

    seqs_for_hmmer_dir = run_log_dir + "/Record4_get_Seqs"

    make_dir(seqs_for_hmmer_dir)

    seq_with_ltrs = seqs_for_hmmer_dir + "/seq_with_paired_ltrs"

    make_dir(seq_with_ltrs)

    seq_with_ltrs_nt = (seq_with_ltrs
                                 + "/seq_with_paired_ltrs.fas")

    seq_with_ltrs_aa = (seq_with_ltrs
                                 + "/seq_with_paired_ltrs_aa.fas")

    flank_without_ltrs = seqs_for_hmmer_dir + "/flank_without_paired_ltrs"

    make_dir(flank_without_ltrs)

    flank_without_ltrs_nt = (flank_without_ltrs
                              + "/flank_without_ltrs.fasta")

    flank_without_ltrs_aa = (flank_without_ltrs
                              + "/flank_without_ltrs_AA.fasta")

    simplify_gff3_dir = identy_pairLTR + "/Simplify_gff3"
    simplify_gff3_path = simplify_gff3_dir + "/harvest_gff3_simplify.txt"

    gff_duplicate_removal_path = (simplify_gff3_dir
                                  + "/harvest_gff3_duplicate_removal.txt")

    LTRharvest_out_size = os.path.getsize(ltrharvest_out_seq)


    if LTRharvest_out_size != 0:

        make_dir(simplify_gff3_dir)

        simplify_gff3_process = SimplyGFF(ltrharvest_gff3_file,
                                           simplify_gff3_path,
                                          gff_duplicate_removal_path,
                                          parameter_dic["mintsd"])
        simplify_gff3_process.run()

        # get sequences for HMMER
        seq_for_hmmer = GetSeqForHmmer(flank_seq_file, simplify_gff3_path,
                                       gff_duplicate_removal_path,
                                       seq_with_ltrs_nt,seq_with_ltrs_aa,
                                       flank_without_ltrs_nt,flank_without_ltrs_aa)
        seq_for_hmmer.run()

    else:
        print("\n" +
            "*** Note: no pair-LTR ERV was found, and the HMMER search domain was started!")

        make_dir(flank_without_ltrs)

        GetSeqForHmmer2(flank_seq_file, flank_without_ltrs_nt,
                        flank_without_ltrs_aa)


    ###  ***Step 5 annotation domain***

    # init ERV.profile database

    if os.path.exists(ERVprofiles_path) == False:
        with zipfile.ZipFile(ERVprofiles_zip, "r") as zip_file:
            zip_file.extractall(os.path.dirname(ERVprofiles_zip))

    h3p_path = ERVprofiles_path + ".h3p"

    if os.path.exists(h3p_path) == False:
        hmmpress_commd = (hmmpress_execute + " " + ERVprofiles_path)

        hmmpress_process = subprocess.Popen(hmmpress_commd,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.STDOUT,
                                            universal_newlines=True,
                                            shell=True)

        while True:

            output = hmmpress_process.stdout.readline()
            if hmmpress_process.poll() is not None:
                break

            elif output:
                print(output)

        hmmpress_process.terminate()



    ##  **Step 5.1 annotation of ervs with paired-LTRs (seq_with_ltrs_aa) **

    pairltr_ervs_map_genome = ""

    if LTRharvest_out_size != 0: 

        hmmer_result_dir = run_log_dir + "/Record5_HMMER_analysis"
        make_dir(hmmer_result_dir)

        # * HMMER analysis *

        pairLTR_ERVs_hmmer = hmmer_result_dir + "/seqs_with_paired_LTRs"

        make_dir(pairLTR_ERVs_hmmer)

        seqs_paired_LTRs_hmmer = (pairLTR_ERVs_hmmer + "/"
                                 + parameter_dic["out_prefix"]
                                 + "_seq_with_paired_LTR_hmmer.txt")

        seqs_paired_LTRs_hmmer_log = (pairLTR_ERVs_hmmer + "/"
                                  + parameter_dic["out_prefix"]
                                  + "_seq_with_paired_LTR_hmmer_log.txt")


        hmmer_commd1 = (hmmscan_execute + " --domtblout "
                        + seqs_paired_LTRs_hmmer
                        + " -E " + parameter_dic["HmmerE"]
                        + " --cpu " + parameter_dic["thread"] + " "
                        + ERVprofiles_path + " "
                        + seq_with_ltrs_aa
                        + " > " + seqs_paired_LTRs_hmmer_log)

        print("\n" + ">>> Start HMMER search for ERVs with paired-LTRs...")

        hmmer1_process = subprocess.Popen(hmmer_commd1,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT,
                                          universal_newlines=True,
                                          shell=True)

        while True:

            output = hmmer1_process.stdout.readline()
            if hmmer1_process.poll() is not None:
                break

            elif output:
                print(output)

        hmmer1_process.terminate()

        if hmmer_contain(seqs_paired_LTRs_hmmer) == "True":
            print("\n" + ">>> Annotate domain...." + "\n")
            #  * simplify the result of hmmer *
            seqs_paired_LTRs_hmmer_simplify = (pairLTR_ERVs_hmmer + "/"
                                           + parameter_dic["out_prefix"]
                                           + "_seq_with_paired_LTR_hmmer_simplify.txt")

            hmmer_head_pairedLTR = (
                        "##seq name for HMMER" + "\t" + "GAG domain"
                        + "\t" + "DUT domain" + "\t"
                        + "AP domain" + "\t" + "RT domain"
                        + "\t" + "RNaseH domain" + "\t"
                        + "INT domain" + "\t" + "ENV domain"
                        + "\n")

            simp_hmmer1 = SimplyHMMER(seqs_paired_LTRs_hmmer,
                                      seqs_paired_LTRs_hmmer_simplify,
                                      hmmer_head_pairedLTR,parameter_dic)
            simp_hmmer1.run()

            # * domian map to genome *
            annotation_dir = (parameter_dic["out_dir"]
                              + "/Annotatio_and_extract_domain")

            make_dir(annotation_dir)
            pairLTR_ervs_annotation_dir = annotation_dir + "/ERVs_with_paired_LTRs"
            make_dir(pairLTR_ervs_annotation_dir)

            pairltr_ervs_map_genome = (pairLTR_ervs_annotation_dir
                                       + "/" + parameter_dic["out_prefix"]
                                       + "_domain_annotation_in_genome_final.txt")

            domain_in_genome_map1 = PairLTRervMap(seqs_paired_LTRs_hmmer_simplify,
                                                  gff_duplicate_removal_path,
                                                  pairltr_ervs_map_genome)
            domain_in_genome_map1.run()

            #  *  extract the domain of ervs with paired-LTRs  *

            doamin_seq_dir1 = (pairLTR_ervs_annotation_dir + "/"
                               + parameter_dic["out_prefix"]
                               + "_domain_seqs")
            make_dir(doamin_seq_dir1)

            pairLTRerv_seq_path = (pairLTR_ervs_annotation_dir + "/"
                                   + parameter_dic["out_prefix"]
                                   + "_extracted_ERVs_with_paired_LTRs.fas")

            pairLTRs_path = (pairLTR_ervs_annotation_dir + "/"
                             + parameter_dic["out_prefix"]
                             + "_extracted_pairedLTR.fas")

            getpairLTRervs_process = GetPairLTRervs(pairltr_ervs_map_genome,
                                                    parameter_dic["host_path"],
                                                    doamin_seq_dir1,
                                                    pairLTRerv_seq_path,
                                                    pairLTRs_path)

            getpairLTRervs_process.run()

            #  * annotate domain  on the extracted pair of LTR_ervs *

            pairltr_ervs_anno_path = (pairLTR_ervs_annotation_dir
                                 + "/domain_annotation_in_ERVs_with_paired_LTRs.txt")

            pairltr_ervs_annotation = PairLTRErvAnno(pairltr_ervs_map_genome,
                                                     pairltr_ervs_anno_path,
                                                     parameter_dic["out_prefix"])

            pairltr_ervs_annotation.run()

            time_end = time.time()

            print("\n" + ">>> The result was saved in: "
                  + annotation_dir + "\n")

        else:
            print("\n" + "*** Note: no domain was found in the identified potential ERVs with paired-LTRs!")



    ##  ** 5.2 annotation of potential ERVLEs whitout paired-LTRs
    no_pairLTR_flankseqaa_size = os.path.getsize(flank_without_ltrs_aa)

    if no_pairLTR_flankseqaa_size != 0:

        hmmer_result_dir = run_log_dir + "/Record5_HMMER_analysis"
        make_dir(hmmer_result_dir)

        # * HMMER analysis *
        no_LTR_flank_hmmer = hmmer_result_dir + "/flank_without_paired_LTRs"

        make_dir(no_LTR_flank_hmmer)

        no_LTR_flank_hmmer_su = (no_LTR_flank_hmmer + "/"
                                 + parameter_dic["out_prefix"]
                                 + "_flank_without_paired_LTR_hmmer.txt")

        no_LTR_flank_hmmer_log = (no_LTR_flank_hmmer + "/"
                                  + parameter_dic["out_prefix"]
                                  + "_flank_without_paired_LTR_hmmer_log.txt")

        hmmer_commd2 = (hmmscan_execute + " --domtblout "
                        + no_LTR_flank_hmmer_su
                        + " -E " + parameter_dic["HmmerE"]
                        + " --cpu " + parameter_dic["thread"] + " "
                        + ERVprofiles_path + " " + flank_without_ltrs_aa
                        + " > " + no_LTR_flank_hmmer_log)

        print("\n" + ">>> Start HMMER search for ERVLEs without paired-LTRs...")
        hmmer2_process = subprocess.Popen(hmmer_commd2,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT,
                                          universal_newlines=True,
                                          shell=True)

        while True:

            output = hmmer2_process.stdout.readline()
            if hmmer2_process.poll() is not None:
                break

            elif output:
                print(output)


        hmmer2_process.terminate()

        if hmmer_contain(no_LTR_flank_hmmer_su) == "True":
            print("\n" + ">>> Annotate domain...." + "\n")

            # * simplify the result of hmmer *
            no_LTR_flank_hmmer_simplify = (no_LTR_flank_hmmer
                                           + "/" + parameter_dic["out_prefix"]
                                           + "_flank_seq_without_paired_LTR_hmmer_simplify.txt")

            hmmer_head_noLTRs = ("##seq name for HMMER" + "\t"
                                       + "GAG domain"
                                       + "\t" + "DUT domain"
                                       + "\t"
                                       + "AP domain" + "\t"
                                       + "RT domain"
                                       + "\t" + "RNaseH domain"
                                       + "\t"
                                       + "INT domain" + "\t"
                                       + "ENV domain"
                                       + "\n")

            simp_hmmer2 = SimplyHMMER(no_LTR_flank_hmmer_su,
                                      no_LTR_flank_hmmer_simplify,
                                      hmmer_head_noLTRs,parameter_dic)
            simp_hmmer2.run()

            # * domian map to genome *
            annotation_dir = (parameter_dic["out_dir"]
                              + "/Annotatio_and_extract_domain")
            make_dir(annotation_dir)

            noLTR_flank_annotation_dir = (annotation_dir
                                          + "/ERVLEs_without_paired_LTRs")

            make_dir(noLTR_flank_annotation_dir)

            noLTR_flank_map_genome = (noLTR_flank_annotation_dir + "/"
                                      + parameter_dic["out_prefix"]
                                      + "_domain_annotation_in_genome.txt")

            out_map_duplicate_removal = (noLTR_flank_annotation_dir + "/"
                                         + parameter_dic["out_prefix"]
                                         + "_domain_annotation_in_genome_final.txt")

            if LTRharvest_out_size != 0:

                domain_in_genome_map2 = NoPairLTReveMap(pairltr_ervs_map_genome,
                                                        no_LTR_flank_hmmer_simplify,
                                                        noLTR_flank_map_genome,
                                                        out_map_duplicate_removal)
                domain_in_genome_map2.run()

            elif LTRharvest_out_size == 0:
                domain_in_genome_map2 = NoPairLTReveMap2(no_LTR_flank_hmmer_simplify,
                                                        noLTR_flank_map_genome,
                                                        out_map_duplicate_removal)

                domain_in_genome_map2.run()


            #  *  extract the potential ERVs ervs and domain in frank sequence *

            doamin_seq_dir2 = (noLTR_flank_annotation_dir + "/"
                               + parameter_dic["out_prefix"] + "_domain_seqs")

            make_dir(doamin_seq_dir2)
            potential_erv_path = (noLTR_flank_annotation_dir + "/"
                                  + parameter_dic["out_prefix"]
                                  + "_extracted_potential_ERVLEs.fas")

            get_potential_erv = GetPotentialEves(out_map_duplicate_removal,
                                                 parameter_dic["host_path"],
                                                 doamin_seq_dir2,
                                                 potential_erv_path)

            get_potential_erv.run()


            #  * annotate domain on the potential ervs *

            domain_in_potential_erv = (noLTR_flank_annotation_dir
                                    + "/domain_annotation_in_potential_ERVLEs.txt")

            potential_ervs_annotation = PotentialEveAnno(out_map_duplicate_removal,
                                                         domain_in_potential_erv,
                                                         parameter_dic["out_prefix"])
            potential_ervs_annotation.run()


            print("\n" + ">>> The result was saved in: " + annotation_dir + "\n")

        else:

            print("\n" + "*** Note: no reliable domain of retroviridae was found in the "
                         "these potential ERVLEs without paired-LTRs!")



    print("\n" + ">>> Analysis have finshed.")


    duration = datetime.today().now() - time_start

    # end_memory = calculate_memory()
    #
    # used_memory = end_memory - start_memory
    #
    # print(">>> " + f"Occupied {used_memory}MB memory in total" + "\n")

    print(">>> " + "Take " + str(duration) + " seconds in total." + "\n")

    sys.exit()


if __name__ == "__main__":
    starts()




# -*- coding: utf-8 -*-

"""
Author: Zhou Zhi-Jian
Email: zjzhou@hnu.edu.cn
Time: 2021/11/11 11:14

"""
import os

from public_function import domain_locate_genome

def domain_start_end(domian):


    if domian.find("complement") == -1:  # 正链上

        domian_start = int((domian.split("(")[-1]).split("..")[0])

        domian_end = int((domian.split("(")[-1]).split("..")[-1].replace(")", ""))

    else: # 负链上
        domian_start = int(
            (domian.split("(")[-1]).split("..")[-1].replace(")", ""))

        domian_end = int((domian.split("(")[-1]).split("..")[0])


    return [domian_start,domian_end]




class PairLTRervMap(object):
    def __init__(self,hmmer_simplify_path,gff3_final_path, out_file):
        super(PairLTRervMap,self).__init__()
        
        self.hmmer_simplify_path = hmmer_simplify_path
        self.gff3_final_path = gff3_final_path
        self.domain_in_genome = out_file


    def run(self):

        HMMER_simplify_file = open(self.hmmer_simplify_path, "r",
                                   encoding="utf-8")
        domain_in_genome_file = open(self.domain_in_genome, "w",encoding="utf-8")

        domain_in_genome_file.write(
            "##Source of ERVs with paired_LTRs "
            + "[host|start|end|chain]"
            + "\t" + "5'-LTR in genome" + "\t" + "3'-LTR in genome"
            + "\t" + "LTR_similarity" + "\t" + "GAG in genome" + "\t"
            + "DUT in genome" + "\t" + "AP in genome" + "\t"
            + "RT in genome" + "\t" + "RNaseH in genome" + "\t"
            + "INT in genome" + "\t" + "ENV in genome" + "\n")


        gff3_final_file = open(self.gff3_final_path, "r",encoding="utf-8")

        ltr_dic = {}

        for line in gff3_final_file:
            line = line.strip()
            if not line.startswith("#"):
                line_list = line.split("\t")
                pair_ltr_name = line_list[-3]
                ltrs = (line_list[-2] + "\t" + line_list[-1]
                        + "\t" + line_list[2])
                ltr_dic[pair_ltr_name] = ltrs

        gff3_final_file.close()

        domain_in_genome_dic = {}

        name_list = []

        for line in HMMER_simplify_file:
            line = line.strip()

            if not line.startswith("##"):
                line_list = line.split("\t")

                seq_name = line_list[0]

                ltrs = ltr_dic[seq_name]

                ltr1, ltr2, ltr_similarary = ltrs.split("\t")

                new_domain_list = []

                same_strand_count = 0
                complement_strand_count = 0

                for n in range(1, len(line_list)):
                    domain_record = line_list[n]
                    if domain_record != "not found" and domain_record.find("relatively short") == -1:
                        (domain_name, c_Evalues, orfs,
                         domain_aa_site,
                         domain_nts_lenth) = domain_record.split(",", 4)

                        seq_coordinate = (str(seq_name.split("|")[1]) + ".."
                                          + str(seq_name.split("|")[2]))

                        seq_stand = seq_name.split("|")[-1]

                        re_strand, re_domain_location = domain_locate_genome(
                            seq_coordinate, seq_stand, domain_aa_site, orfs)

                        new_domain = (domain_name + "," + c_Evalues + ","
                                      + re_domain_location)

                        if re_strand == "same":
                            same_strand_count += 1
                        elif re_strand == "complement":
                            complement_strand_count += 1

                    else:
                        new_domain = domain_record

                    new_domain_list.append(new_domain)

                if same_strand_count >= complement_strand_count:
                    finnaly_strand = "same"

                    ltr1_start = int(ltr1.split("..")[0])
                    ltr2_start = int(ltr2.split("..")[0])

                    if ltr1_start < ltr2_start:
                        five_ltr = ltr1  # Take the smallest value as 5’-LTR
                        three_ltr = ltr2

                    else:
                        five_ltr = ltr2  # Take the smallest value as 5’-LTR
                        three_ltr = ltr1


                else:
                    finnaly_strand = "complement"

                    ltr1_start = int(ltr1.split("..")[0])
                    ltr2_start = int(ltr2.split("..")[0])

                    if ltr1_start < ltr2_start:
                        # The complementary chain takes the larger value as 5’-LTR
                        five_ltr = "complement(" + ltr2 + ")"
                        three_ltr = "complement(" + ltr1 + ")"

                    else:
                        five_ltr = "complement(" + ltr1 + ")"
                        three_ltr = "complement(" + ltr2 + ")"


                name_ = seq_name.replace(seq_name.split("|")[-1],"").strip("|") + "|" + finnaly_strand

                new_line = ( name_ + "\t" + five_ltr
                            + "\t" + three_ltr
                            + "\t" + ltr_similarary + "\t"
                            + "\t".join(new_domain_list))

                name_list.append(name_)
                domain_in_genome_dic[name_] = new_line




        use_list = []
        for line1 in name_list:

            qury_name, qury_start, qury_end, qury_direction = line1.split("|",
                                                                          3)

            for line2 in name_list:
                if line1 != line2:

                    (refer_name, refer_start,
                     refer_end, refer_direction) = line2.split("|", 3)
                    if (qury_name == refer_name
                        and qury_direction) == refer_direction:
                        if qury_start >= refer_start and qury_end <= refer_end:
                            break

            else:
                use_list.append(line1)




        woqu_ = {}

        finnaly_use = []

        for each_name in use_list:
            erv_contain_list = domain_in_genome_dic[each_name].split("\t")  

            xxx_list = []

            bat_seq_index = erv_contain_list[0].split("|")[0]  

            ltrs_similary = float(erv_contain_list[3]) 

            xxx_list.append(bat_seq_index)

            for n in range(4,len(erv_contain_list)):

                domain_ = erv_contain_list[n]

                if domain_ != "not found" and domain_.find("relatively short") == -1:
                    xxx_list.append(domain_.split(",")[-1]) 

                else:

                    xxx_list.append(domain_)

            label_ = "\t".join(xxx_list) 

            if not label_ in woqu_: 
                finnaly_use.append(each_name) 
                woqu_[label_] = [each_name,ltrs_similary]

            else:
                if ltrs_similary > woqu_[label_][1]: 
                    finnaly_use.remove(woqu_[label_][0])
                    finnaly_use.append(each_name) 
                    woqu_[label_] = [each_name, ltrs_similary] 


        for erv_name in finnaly_use:

            flage = 0

            poss_erv = domain_in_genome_dic[erv_name]

            poss_erv_contain_list = poss_erv.split("\t")

            erv_name = poss_erv_contain_list[0]

            chain_direction = erv_name.split("|")[-1]

            GAG_ = poss_erv_contain_list[4].strip()

            PRO_ = poss_erv_contain_list[6].strip()

            RT_ = poss_erv_contain_list[7].strip()

            RH_ = poss_erv_contain_list[8].strip()

            INT_ = poss_erv_contain_list[9].strip()

            ENV_ = poss_erv_contain_list[-1].strip()

            if GAG_.find(",") != -1 and RT_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)

                RT_site = domain_start_end(RT_)

                if chain_direction == "same" and RT_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and RT_site[0] > Gag_site[
                    0]:

                    flage = 1

            if GAG_.find(",") != -1 and RH_.find(",") != -1:
                Gag_site = domain_start_end(GAG_)
                RH_site = domain_start_end(RH_)

                if chain_direction == "same" and RH_site[0] < Gag_site[0]:
                    flage = 1

                elif chain_direction == "complement" and RH_site[0] > Gag_site[
                    0]:
                    flage = 1



            if GAG_.find(",") != -1 and INT_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)

                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and INT_site[0] > Gag_site[0]:

                    flage = 1

            if GAG_.find(",") != -1 and ENV_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)
                ENV_site = domain_start_end(ENV_)

                if chain_direction == "same" and ENV_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and ENV_site[0] > Gag_site[
                    0]:

                    flage = 1

            if RT_.find(",") != -1 and ENV_.find(",") != -1:
                RT_site = domain_start_end(RT_)
                ENV_site = domain_start_end(ENV_)

                if chain_direction == "same" and ENV_site[0] < RT_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > RT_site[
                    0]:

                    flage = 1


            if RT_.find(",") != -1 and RH_.find(",") != -1:
                RT_site = domain_start_end(RT_)
                RH_site = domain_start_end(RH_)

                if chain_direction == "same" and RT_site[0] > RH_site[0]: 
                    flage = 1

                elif chain_direction == "complement" and RT_site[0] < RH_site[0]:
                    flage = 1

                if abs(RT_site[0] - RH_site[0]) > 5000: 
                    flage = 1


            if RT_.find(",") != -1 and INT_.find(",") != -1:


                RT_site = domain_start_end(RT_)
                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < RT_site[0]:  
                    flage = 1

                elif chain_direction == "complement" and INT_site[0] > RT_site[0]:
                    flage = 1


            if RH_.find(",") != -1 and INT_.find(",") != -1:
                RH_site =  domain_start_end(RH_)
                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < RH_site[0]:  
                    flage = 1

                elif chain_direction == "complement" and INT_site[0] > RH_site[0]:
                    flage = 1



            if PRO_.find(",") != -1 and RT_.find(",") != -1:
                PRO_site = domain_start_end(PRO_)
                RT_site = domain_start_end(RT_)

                if chain_direction == "same" and RT_site[0] < PRO_site[0]:
                    flage = 1

                elif chain_direction == "complement" and RT_site[0] > PRO_site[0]:
                    flage = 1


            if PRO_.find(",") != -1 and ENV_.find(",") != -1:
                PRO_site = domain_start_end(PRO_)

                ENV_site = domain_start_end(ENV_)

                if chain_direction == "same" and ENV_site[0] < PRO_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > PRO_site[
                    0]:

                    flage = 1

            if INT_.find(",") != -1 and ENV_.find(",") != -1:
                INT_site = domain_start_end(INT_)

                ENV_site = domain_start_end(ENV_)

                if chain_direction == "same" and ENV_site[0] < INT_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > INT_site[
                    0]:

                    flage = 1

            if flage == 0:


                domain_in_genome_file.write(domain_in_genome_dic[erv_name] + "\n")


        domain_in_genome_file.close()




class NoPairLTReveMap(object):
    def __init__(self, pairltr_ervs_map_genome,hmmer_simplify_path, out_map,
                 out_map_duplicate_removal):
        super(NoPairLTReveMap,self).__init__()

        self.pairltr_ervs_map_genome = pairltr_ervs_map_genome
        self.hmmer_simplify_path = hmmer_simplify_path
        self.domain_in_genome = out_map
        self.out_map_duplicate_removal = out_map_duplicate_removal


    def run(self):

        HMMER_simplify_file = open(self.hmmer_simplify_path, "r",encoding="utf-8")

        domain_in_genome_file = open(self.domain_in_genome, "w",encoding="utf-8")
        domain_in_genome_file.write(
            "##flank sequence used for HMMER"
            + "\t"
            + "potential ERVLEs without paired-LTRs "
            + "[host|start|end|chain]"
            + "\t" + "GAG in genome" + "\t"
            + "DUT in genome" + "\t" + "AP in genome" + "\t"
            + "RT in genome" + "\t" + "RNaseH in genome" + "\t"
            + "INT in genome" + "\t" + "ENV in genome" + "\n")



        for line in HMMER_simplify_file:
            line = line.strip()
            if not line.startswith("##"):
                line_list = line.split("\t")

                seq_name = line_list[0]

                new_domain_list = []

                same_strand_count = 0
                complement_strand_count = 0

                max_site_list = []
                min_site_list = []

                for n in range(1, len(line_list)):

                    domain_record = line_list[n]

                    if domain_record != "not found" and domain_record.find("relatively short") == -1:
                        (domain_name, c_Evalues, orfs,
                         domain_aa_site,
                         domain_nts_lenth) = domain_record.split(",", 4)

                        seq_coordinate = (str(seq_name.split("|")[1]) + ".."
                                          + str(seq_name.split("|")[2]))

                        seq_stand = seq_name.split("|")[-1]

                        re_strand, re_domain_location = domain_locate_genome(
                            seq_coordinate, seq_stand, domain_aa_site, orfs)

                        new_domain = (domain_name + "," + c_Evalues + ","
                                      + re_domain_location)

                        domain_range = \
                        (re_domain_location.split("(")[-1]).split(")")[0]
                        domain_start = int(domain_range.split("..")[0])
                        domain_end = int(domain_range.split("..")[1])

                        max_site_list.append(domain_end)
                        min_site_list.append(domain_start)

                        if re_strand == "same":
                            same_strand_count += 1
                        elif re_strand == "complement":
                            complement_strand_count += 1

                    else:
                        new_domain = domain_record

                    new_domain_list.append(new_domain)

                if same_strand_count >= complement_strand_count:
                    finnaly_strand = "same"

                else:
                    finnaly_strand = "complement"

                max_site_list.sort()
                min_site_list.sort()

                max_site = max_site_list[-1]
                min_site = min_site_list[0]

                if abs(max_site - min_site) <= 12000: 

                    new_line = (seq_name + "\t"
                                + seq_name.split("|")[0] + "|"
                                + str(min_site) + "|" + str(max_site)
                                + "|" + finnaly_strand + "\t"
                                + "\t".join(new_domain_list) + "\n")

                    domain_in_genome_file.write(new_line)

        domain_in_genome_file.close()


        """
        Remove potential regions of erv that overlap with 
        the scope after relocation 
        """
        domain_in_genome_file = open(self.domain_in_genome, "r",
                                     encoding="utf-8")

        old_seq_name = []

        ervles_dic = {}
        use_list = []

        for line in domain_in_genome_file:
            if not line.startswith("##"):
                seq_name = line.split("\t")[1]
                old_seq_name.append(seq_name.strip())

                ervles_dic[seq_name] = (line.split("\t")[1] + "\t" + 
                                      line.split("\t")[2] + "\t" +
                                      line.split("\t")[3] + "\t" +
                                      line.split("\t")[4] + "\t" +
                                      line.split("\t")[5] + "\t" +
                                      line.split("\t")[6] + "\t" +
                                      line.split("\t")[7] + "\t" +
                                      line.split("\t")[8]).strip()


        domain_in_genome_file.close()

        os.remove(self.domain_in_genome)



        for line1 in old_seq_name:

            qury_name, qury_start, qury_end, qury_direction = line1.split("|",
                                                                          3)

            for line2 in old_seq_name:
                if line1 != line2:

                    (refer_name, refer_start,
                     refer_end, refer_direction) = line2.split("|", 3)
                    if (qury_name == refer_name
                        and qury_direction) == refer_direction:
                        if qury_start >= refer_start and qury_end <= refer_end:
                            break

            else:
                use_list.append(line1)


        ganjing = set()

        finnaly_use1 = []

        for each_seqname in use_list:
            if not each_seqname in ganjing:
                finnaly_use1.append(each_seqname)
                ganjing.add(each_seqname)


  
        woqu_ = set()

        finnaly_use2 = []

        for each_seqname in finnaly_use1:
            erv_contain_list = ervles_dic[each_seqname].split("\t") 

            xxx_list = []

            bat_seq_index = erv_contain_list[0].split("|")[0] 

            xxx_list.append(bat_seq_index)

            for n in range(1,len(erv_contain_list)):

                domain_ = erv_contain_list[n]

                if domain_ != "not found" and domain_.find("relatively short") == -1:
                    xxx_list.append(domain_.split(",")[-1]) 

                else:

                    xxx_list.append(domain_)

            label_ = "\t".join(xxx_list)

            if not label_ in woqu_:
                finnaly_use2.append(each_seqname)
                woqu_.add(label_)



        ERV_anno_domian_dic = {}

        with open(self.pairltr_ervs_map_genome,"r",encoding="utf-8") as pairltr_ervs_anno:

            for line in pairltr_ervs_anno:

                if not line.startswith("##"):

                    line = line.strip()

                    line_list = line.split("\t")

                    key = line_list[0].split("|")[0] + "_" + \
                          line_list[0].split("|")[-1]

                    if key not in ERV_anno_domian_dic:
                        ERV_anno_domian_dic[key] = []

                    for n in range(4, len(line_list)):
                        domain = line_list[n]
                        if domain.find("relatively short") == -1 and domain != "not found":
                            ERV_anno_domian_dic[key].append(domain.split(",")[-1])



        finnaly_use3 = []

        for ervle_name in finnaly_use2:

            flage = 0

            line = ervles_dic[ervle_name]

            line = line.strip()

            line_list = line.split("\t")

            key = line_list[0].split("|")[0] + "_" + line_list[0].split("|")[-1]


            for n in range(1, len(line_list)):
                domain = line_list[n]

                if domain.find("relatively short") == -1 and domain != "not found":

                    local_ = domain.split(",")[-1]

                    if key in ERV_anno_domian_dic:

                        ERV_anno_domian_list = ERV_anno_domian_dic[key]

                        if local_ in ERV_anno_domian_list:

                            flage = 1

                            break

            if flage == 0:
                finnaly_use3.append(ervle_name)


        domain_duplicate_removal = open(self.out_map_duplicate_removal, "w",
                                        encoding="utf-8")

        domain_duplicate_removal.write(
            "##potential ERVLEs without paired-LTRs "
            + "[host|start|end|chain]"
            + "\t" + "GAG in genome" + "\t"
            + "DUT in genome" + "\t" + "AP in genome" + "\t"
            + "RT in genome" + "\t" + "RNaseH in genome" + "\t"
            + "INT in genome" + "\t" + "ENV in genome" + "\n")



        for line in finnaly_use3:

            flage = 0

            poss_erv = ervles_dic[line]

            poss_erv_contain_list = poss_erv.split("\t")

            erv_name = poss_erv_contain_list[0]

            chain_direction = erv_name.split("|")[-1]

            GAG_ = poss_erv_contain_list[1].strip()

            PRO_ = poss_erv_contain_list[3].strip()

            RT_ = poss_erv_contain_list[4].strip()

            RH_ = poss_erv_contain_list[5].strip()

            INT_ = poss_erv_contain_list[6].strip()

            ENV_ = poss_erv_contain_list[-1].strip()


            if GAG_.find(",") != -1 and RT_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)

                RT_site = domain_start_end(RT_)

                if chain_direction == "same" and RT_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and RT_site[0] > Gag_site[0]:

                    flage = 1

            if GAG_.find(",") != -1 and RH_.find(",") != -1:
                Gag_site = domain_start_end(GAG_)
                RH_site = domain_start_end(RH_)

                if chain_direction == "same" and RH_site[0] < Gag_site[0]:
                    flage = 1

                elif chain_direction == "complement" and RH_site[0] > Gag_site[0]:
                    flage = 1


            if GAG_.find(",") != -1 and INT_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)

                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and INT_site[0] > Gag_site[0]:

                    flage = 1



            if GAG_.find(",") != -1 and ENV_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)
                ENV_site = domain_start_end(ENV_)

                if chain_direction == "same" and ENV_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and ENV_site[0] > Gag_site[0]:

                    flage = 1


                if abs(Gag_site[0] - ENV_site[0]) > 12000:  
                    flage = 1




            if RT_.find(",") != -1 and ENV_.find(",") != -1:
                RT_site = domain_start_end(RT_)
                ENV_site = domain_start_end(ENV_)

                if chain_direction == "same" and ENV_site[0] < RT_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > RT_site[0]:

                    flage = 1

                if abs(RT_site[0] - ENV_site[0]) > 6000:  
                    flage = 1



            if PRO_.find(",") != -1 and RT_.find(",") != -1:
                PRO_site = domain_start_end(PRO_)
                RT_site = domain_start_end(RT_)

                if chain_direction == "same" and RT_site[0] < PRO_site[0]:
                    flage = 1

                elif chain_direction == "complement" and RT_site[0] > PRO_site[0]:
                    flage = 1



            if RT_.find(",") != -1 and INT_.find(",") != -1:


                RT_site = domain_start_end(RT_)
                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < RT_site[
                    0]:  
                    flage = 1

                elif chain_direction == "complement" and INT_site[0] > RT_site[0]:
                    flage = 1


            if RH_.find(",") != -1 and INT_.find(",") != -1:
                RH_site =  domain_start_end(RH_)
                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < RH_site[0]: 
                    flage = 1

                elif chain_direction == "complement" and INT_site[0] > RH_site[0]:
                    flage = 1



            if RT_.find(",") != -1 and RH_.find(",") != -1:
                RT_site = domain_start_end(RT_)
                RH_site = domain_start_end(RH_)

                if chain_direction == "same" and RT_site[0] > RH_site[0]: 
                    flage = 1

                elif chain_direction == "complement" and RT_site[0] < RH_site[0]:
                    flage = 1


                if abs(RT_site[0] - RH_site[0]) > 5000: 
                    flage = 1


            if PRO_.find(",") != -1 and ENV_.find(",") != -1:
                PRO_site = domain_start_end(PRO_)

                ENV_site = domain_start_end(ENV_)


                if chain_direction == "same" and ENV_site[0] < PRO_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > PRO_site[0]:

                    flage = 1


            if INT_.find(",") != -1 and ENV_.find(",") != -1:
                INT_site = domain_start_end(INT_)

                ENV_site = domain_start_end(ENV_)


                if chain_direction == "same" and ENV_site[0] < INT_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > INT_site[0]:

                    flage = 1


            if flage == 0:

                domain_duplicate_removal.write(ervles_dic[erv_name] + "\n")

        domain_duplicate_removal.close()
        





class NoPairLTReveMap2(object):

    """
    if not ERV with paired LTRs in genome

    """

    def __init__(self, hmmer_simplify_path, out_map,out_map_duplicate_removal):
        super(NoPairLTReveMap2,self).__init__()

        self.hmmer_simplify_path = hmmer_simplify_path
        self.domain_in_genome = out_map
        self.out_map_duplicate_removal = out_map_duplicate_removal


    def run(self):

        HMMER_simplify_file = open(self.hmmer_simplify_path, "r",encoding="utf-8")

        domain_in_genome_file = open(self.domain_in_genome, "w",encoding="utf-8")
        domain_in_genome_file.write(
            "##flank sequence used for HMMER"
            + "\t"
            + "potential ERVLEs without paired-LTRs "
            + "[host|start|end|chain]"
            + "\t" + "GAG in genome" + "\t"
            + "DUT in genome" + "\t" + "AP in genome" + "\t"
            + "RT in genome" + "\t" + "RNaseH in genome" + "\t"
            + "INT in genome" + "\t" + "ENV in genome" + "\n")



        for line in HMMER_simplify_file:
            line = line.strip()
            if not line.startswith("##"):
                line_list = line.split("\t")

                seq_name = line_list[0]

                new_domain_list = []

                same_strand_count = 0
                complement_strand_count = 0

                max_site_list = []
                min_site_list = []

                for n in range(1, len(line_list)):

                    domain_record = line_list[n]

                    if domain_record != "not found" and domain_record.find("relatively short") == -1:
                        (domain_name, c_Evalues, orfs,
                         domain_aa_site,
                         domain_nts_lenth) = domain_record.split(",", 4)

                        seq_coordinate = (str(seq_name.split("|")[1]) + ".."
                                          + str(seq_name.split("|")[2]))

                        seq_stand = seq_name.split("|")[-1]

                        re_strand, re_domain_location = domain_locate_genome(
                            seq_coordinate, seq_stand, domain_aa_site, orfs)

                        new_domain = (domain_name + "," + c_Evalues + ","
                                      + re_domain_location)

                        domain_range = \
                        (re_domain_location.split("(")[-1]).split(")")[0]
                        domain_start = int(domain_range.split("..")[0])
                        domain_end = int(domain_range.split("..")[1])

                        max_site_list.append(domain_end)
                        min_site_list.append(domain_start)

                        if re_strand == "same":
                            same_strand_count += 1
                        elif re_strand == "complement":
                            complement_strand_count += 1

                    else:
                        new_domain = domain_record

                    new_domain_list.append(new_domain)

                if same_strand_count >= complement_strand_count:
                    finnaly_strand = "same"

                else:
                    finnaly_strand = "complement"

                max_site_list.sort()
                min_site_list.sort()

                max_site = max_site_list[-1]
                min_site = min_site_list[0]

                if abs(max_site - min_site) <= 12000:

                    new_line = (seq_name + "\t"
                                + seq_name.split("|")[0] + "|"
                                + str(min_site) + "|" + str(max_site)
                                + "|" + finnaly_strand + "\t"
                                + "\t".join(new_domain_list) + "\n")

                    domain_in_genome_file.write(new_line)

        domain_in_genome_file.close()


        """
        Remove potential regions of erv that overlap with 
        the scope after relocation 
        """
        domain_in_genome_file = open(self.domain_in_genome, "r",
                                     encoding="utf-8")

        old_seq_name = []

        ervles_dic = {}
        use_list = []

        for line in domain_in_genome_file:
            if not line.startswith("##"):
                seq_name = line.split("\t")[1]
                old_seq_name.append(seq_name.strip())

                ervles_dic[seq_name] = (line.split("\t")[1] + "\t" +
                                      line.split("\t")[2] + "\t" +
                                      line.split("\t")[3] + "\t" +
                                      line.split("\t")[4] + "\t" +
                                      line.split("\t")[5] + "\t" +
                                      line.split("\t")[6] + "\t" +
                                      line.split("\t")[7] + "\t" +
                                      line.split("\t")[8]).strip()


        domain_in_genome_file.close()

        os.remove(self.domain_in_genome)


        for line1 in old_seq_name:

            qury_name, qury_start, qury_end, qury_direction = line1.split("|",
                                                                          3)

            for line2 in old_seq_name:
                if line1 != line2:

                    (refer_name, refer_start,
                     refer_end, refer_direction) = line2.split("|", 3)
                    if (qury_name == refer_name
                        and qury_direction) == refer_direction:
                        if qury_start >= refer_start and qury_end <= refer_end:
                            break

            else:
                use_list.append(line1)


        ganjing = set()

        finnaly_use1 = []

        for each_seqname in use_list:
            if not each_seqname in ganjing:
                finnaly_use1.append(each_seqname)
                ganjing.add(each_seqname)


        woqu_ = set()

        finnaly_use2 = []

        for each_seqname in finnaly_use1:
            erv_contain_list = ervles_dic[each_seqname].split("\t") 

            xxx_list = []

            bat_seq_index = erv_contain_list[0].split("|")[0]

            xxx_list.append(bat_seq_index)

            for n in range(1,len(erv_contain_list)):

                domain_ = erv_contain_list[n]

                if domain_ != "not found" and domain_.find("relatively short") == -1:
                    xxx_list.append(domain_.split(",")[-1]) 

                else:

                    xxx_list.append(domain_)

            label_ = "\t".join(xxx_list)

            if not label_ in woqu_:
                finnaly_use2.append(each_seqname)
                woqu_.add(label_)




        domain_duplicate_removal = open(self.out_map_duplicate_removal, "w",
                                        encoding="utf-8")

        domain_duplicate_removal.write(
            "##potential ERVLEs without paired-LTRs "
            + "[host|start|end|chain]"
            + "\t" + "GAG in genome" + "\t"
            + "DUT in genome" + "\t" + "AP in genome" + "\t"
            + "RT in genome" + "\t" + "RNaseH in genome" + "\t"
            + "INT in genome" + "\t" + "ENV in genome" + "\n")



        for line in finnaly_use2:

            flage = 0

            poss_erv = ervles_dic[line]

            poss_erv_contain_list = poss_erv.split("\t")

            erv_name = poss_erv_contain_list[0]

            chain_direction = erv_name.split("|")[-1]

            GAG_ = poss_erv_contain_list[1].strip()

            PRO_ = poss_erv_contain_list[3].strip()

            RT_ = poss_erv_contain_list[4].strip()

            RH_ = poss_erv_contain_list[5].strip()

            INT_ = poss_erv_contain_list[6].strip()

            ENV_ = poss_erv_contain_list[-1].strip()


            if GAG_.find(",") != -1 and RT_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)

                RT_site = domain_start_end(RT_)

                if chain_direction == "same" and RT_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and RT_site[0] > Gag_site[0]:

                    flage = 1

            if GAG_.find(",") != -1 and RH_.find(",") != -1:
                Gag_site = domain_start_end(GAG_)
                RH_site = domain_start_end(RH_)

                if chain_direction == "same" and RH_site[0] < Gag_site[0]:
                    flage = 1

                elif chain_direction == "complement" and RH_site[0] > Gag_site[0]:
                    flage = 1


            if GAG_.find(",") != -1 and INT_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)

                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and INT_site[0] > Gag_site[0]:

                    flage = 1



            if GAG_.find(",") != -1 and ENV_.find(",") != -1:

                Gag_site = domain_start_end(GAG_)
                ENV_site = domain_start_end(ENV_)

                if chain_direction == "same" and ENV_site[0] < Gag_site[0]:
                    flage = 1


                elif chain_direction == "complement" and ENV_site[0] > Gag_site[0]:

                    flage = 1


                if abs(Gag_site[0] - ENV_site[0]) > 12000:
                    flage = 1




            if RT_.find(",") != -1 and ENV_.find(",") != -1:
                RT_site = domain_start_end(RT_)
                ENV_site = domain_start_end(ENV_)

                if chain_direction == "same" and ENV_site[0] < RT_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > RT_site[0]:

                    flage = 1

                if abs(RT_site[0] - ENV_site[0]) > 6000:
                    flage = 1



            if PRO_.find(",") != -1 and RT_.find(",") != -1:
                PRO_site = domain_start_end(PRO_)
                RT_site = domain_start_end(RT_)

                if chain_direction == "same" and RT_site[0] < PRO_site[0]:
                    flage = 1

                elif chain_direction == "complement" and RT_site[0] > PRO_site[0]:
                    flage = 1



            if RT_.find(",") != -1 and INT_.find(",") != -1:

                RT_site = domain_start_end(RT_)
                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < RT_site[
                    0]:
                    flage = 1

                elif chain_direction == "complement" and INT_site[0] > RT_site[0]:
                    flage = 1


            if RH_.find(",") != -1 and INT_.find(",") != -1:
                RH_site =  domain_start_end(RH_)
                INT_site = domain_start_end(INT_)

                if chain_direction == "same" and INT_site[0] < RH_site[0]:
                    flage = 1

                elif chain_direction == "complement" and INT_site[0] > RH_site[0]:
                    flage = 1



            if RT_.find(",") != -1 and RH_.find(",") != -1:
                RT_site = domain_start_end(RT_)
                RH_site = domain_start_end(RH_)

                if chain_direction == "same" and RT_site[0] > RH_site[0]: 
                    flage = 1

                elif chain_direction == "complement" and RT_site[0] < RH_site[0]:
                    flage = 1


                if abs(RT_site[0] - RH_site[0]) > 5000:
                    flage = 1


            if PRO_.find(",") != -1 and ENV_.find(",") != -1:
                PRO_site = domain_start_end(PRO_)

                ENV_site = domain_start_end(ENV_)


                if chain_direction == "same" and ENV_site[0] < PRO_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > PRO_site[0]:

                    flage = 1


            if INT_.find(",") != -1 and ENV_.find(",") != -1:
                INT_site = domain_start_end(INT_)

                ENV_site = domain_start_end(ENV_)


                if chain_direction == "same" and ENV_site[0] < INT_site[0]:

                    flage = 1

                elif chain_direction == "complement" and ENV_site[0] > INT_site[0]:

                    flage = 1


            if flage == 0:

                domain_duplicate_removal.write(ervles_dic[erv_name] + "\n")

        domain_duplicate_removal.close()

